@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2021 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Team
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2016-2018 Mutnick <mutnick@techie.com>
 # COPYRIGHT (C) 2008-2011 Quinox <quinox@users.sf.net>
@@ -55,11 +55,28 @@ class Config:
         config_dir, self.data_dir = self.get_user_directories()
         self.filename = os.path.join(config_dir, "config")
         self.plugin_dir = os.path.join(self.data_dir, "plugins")
-        self.version = "3.1.1"
+        self.version = "3.2.1"
+        self.python_version = sys.version
+        self.gtk_version = ""
 
+        self.application_name = "Nicotine+"
+        self.application_id = "org.nicotine_plus.Nicotine"
+        self.summary = _("Graphical client for the Soulseek peer-to-peer network")
+        self.copyright = """© 2004–2022 Nicotine+ Team
+© 2003–2004 Nicotine Team
+© 2001–2003 PySoulSeek Contributors"""
+
+        self.website_url = "https://nicotine-plus.org/"
+        self.privileges_url = "https://www.slsknet.org/userlogin.php?username=%s"
+        self.portchecker_url = "http://tools.slsknet.org/porttest.php?port=%s"
+        self.issue_tracker_url = "https://github.com/nicotine-plus/nicotine-plus/issues"
+        self.translations_url = "https://nicotine-plus.org/doc/TRANSLATIONS"
+
+        self.config_loaded = False
         self.parser = configparser.ConfigParser(strict=False, interpolation=None)
         self.sections = defaultdict(dict)
         self.defaults = {}
+        self.removed_options = {}
 
     @staticmethod
     def get_user_directories():
@@ -139,10 +156,10 @@ class Config:
         log_dir = os.path.join(self.data_dir, "logs")
         self.defaults = {
             "server": {
-                "server": ('server.slsknet.org', 2242),
-                "login": '',
-                "passw": '',
-                "interface": '',
+                "server": ("server.slsknet.org", 2242),
+                "login": "",
+                "passw": "",
+                "interface": "",
                 "ctcpmsgs": False,
                 "autosearch": [],
                 "autoreply": "",
@@ -161,13 +178,11 @@ class Config:
                 "private_chatrooms": False,
                 "command_aliases": {}
             },
-
             "transfers": {
                 "incompletedir": os.path.join(self.data_dir, 'incomplete'),
                 "downloaddir": os.path.join(self.data_dir, 'downloads'),
                 "uploaddir": os.path.join(self.data_dir, 'received'),
                 "usernamesubfolders": False,
-                "sharedownloaddir": False,
                 "shared": [],
                 "buddyshared": [],
                 "uploadbandwidth": 10,
@@ -192,15 +207,11 @@ class Config:
                 "customgeoblock": "Sorry, your country is blocked",
                 "queuelimit": 10000,
                 "filelimit": 100,
-                "friendsonly": False,
                 "buddysharestrustedonly": False,
                 "friendsnolimits": False,
-                "enablebuddyshares": False,
-                "enabletransferbuttons": True,
                 "groupdownloads": "folder_grouping",
                 "groupuploads": "folder_grouping",
                 "geoblock": False,
-                "geopanic": False,
                 "geoblockcc": [""],
                 "remotedownloads": True,
                 "uploadallowed": 2,
@@ -222,12 +233,10 @@ class Config:
                 "downloadsexpanded": True,
                 "uploadsexpanded": True
             },
-
             "userinfo": {
                 "descr": "''",
                 "pic": ""
             },
-
             "words": {
                 "censored": [],
                 "autoreplaced": {
@@ -253,7 +262,6 @@ class Config:
                 "aliases": True,
                 "onematch": False
             },
-
             "logging": {
                 "debug": False,
                 "debugmodes": [],
@@ -275,27 +283,23 @@ class Config:
                 "readprivatelines": 15,
                 "rooms": []
             },
-
             "privatechat": {
                 "store": True,
                 "users": []
             },
-
             "columns": {
                 "file_search": {},
                 "download": {},
                 "upload": {},
                 "user_browse": {},
                 "buddy_list": {},
-                "chat_room": {},
-                "hideflags": False
+                "chat_room": {}
             },
-
             "searches": {
                 "expand_searches": True,
                 "group_searches": "folder_grouping",
                 "maxresults": 50,
-                "re_filter": False,
+                "enable_history": True,
                 "history": [],
                 "enablefilters": False,
                 "filters_visible": False,
@@ -309,9 +313,9 @@ class Config:
                 "search_results": True,
                 "max_displayed_results": 1500,
                 "min_search_chars": 3,
-                "remove_special_chars": True
+                "remove_special_chars": True,
+                "private_search_results": True
             },
-
             "ui": {
                 "dark_mode": False,
                 "header_bar": True,
@@ -332,7 +336,6 @@ class Config:
                 "inputcolor": "",
                 "spellcheck": True,
                 "exitdialog": 1,
-                "notexists": True,
                 "tab_default": "",
                 "tab_hilite": "#497ec2",
                 "tab_changed": "#497ec2",
@@ -344,22 +347,14 @@ class Config:
                 "tabbrowse": "Top",
                 "tabsearch": "Top",
                 "tab_status_icons": True,
-                "labelmain": 0,
-                "labelrooms": 0,
-                "labelprivate": 0,
-                "labelinfo": 0,
-                "labelbrowse": 0,
-                "labelsearch": 0,
-                "decimalsep": ",",
                 "globalfont": "",
                 "chatfont": "",
-                "roomlistcollapsed": False,
                 "tabclosers": True,
                 "searchfont": "",
                 "listfont": "",
                 "browserfont": "",
                 "transfersfont": "",
-                "last_tab_id": 0,
+                "last_tab_id": "",
                 "modes_visible": {
                     "search": True,
                     "downloads": True,
@@ -381,14 +376,13 @@ class Config:
                     "chatrooms",
                     "interests"
                 ],
-                "showaway": False,
                 "buddylistinchatrooms": "tab",
                 "trayicon": True,
                 "startup_hidden": False,
                 "filemanager": "",
                 "speechenabled": False,
-                "speechprivate": "%(user)s told you.. %(message)s",
-                "speechrooms": "In %(room)s, %(user)s said %(message)s",
+                "speechprivate": "User %(user)s told you: %(message)s",
+                "speechrooms": "In room %(room)s, user %(user)s said: %(message)s",
                 "speechcommand": "flite -t $",
                 "width": 1000,
                 "height": 600,
@@ -396,27 +390,19 @@ class Config:
                 "yposition": -1,
                 "maximized": True,
                 "urgencyhint": True,
-                "private_search_results": False,
-                "private_shares": False,
                 "file_path_tooltips": True,
                 "reverse_file_paths": True
             },
-
             "private_rooms": {
                 "rooms": {}
             },
-
             "urls": {
-                "urlcatching": True,
-                "protocols": {},
-                "humanizeurls": True
+                "protocols": {}
             },
-
             "interests": {
                 "likes": [],
                 "dislikes": []
             },
-
             "players": {
                 "default": "",
                 "npothercommand": "",
@@ -424,11 +410,9 @@ class Config:
                 "npformatlist": [],
                 "npformat": ""
             },
-
             "notifications": {
                 "notification_window_title": True,
                 "notification_tab_colors": False,
-                "notification_tab_icons": True,
                 "notification_popup_sound": False,
                 "notification_popup_file": True,
                 "notification_popup_folder": True,
@@ -436,12 +420,10 @@ class Config:
                 "notification_popup_chatroom": False,
                 "notification_popup_chatroom_mention": True
             },
-
             "plugins": {
                 "enable": True,
                 "enabled": []
             },
-
             "statistics": {
                 "started_downloads": 0,
                 "completed_downloads": 0,
@@ -452,6 +434,114 @@ class Config:
             }
         }
 
+        self.removed_options = {
+            "transfers": (
+                "pmqueueddir",
+                "autoretry_downloads",
+                "shownotification",
+                "shownotificationperfolder",
+                "prioritize",
+                "sharedownloaddir",
+                "geopanic",
+                "enablebuddyshares",
+                "friendsonly",
+                "enabletransferbuttons"
+            ),
+            "server": (
+                "lastportstatuscheck",
+                "serverlist",
+                "enc",
+                "fallbackencodings",
+                "roomencoding",
+                "userencoding",
+                "firewalled"
+            ),
+            "ui": (
+                "enabletrans",
+                "mozembed",
+                "open_in_mozembed",
+                "tooltips",
+                "transalpha",
+                "transfilter",
+                "transtint",
+                "soundenabled",
+                "soundtheme",
+                "soundcommand",
+                "tab_colors",
+                "tab_icons",
+                "searchoffline",
+                "chat_hidebuttons",
+                "tab_reorderable",
+                "private_search_results",
+                "private_shares",
+                "labelmain",
+                "labelrooms",
+                "labelprivate",
+                "labelinfo",
+                "labelbrowse",
+                "labelsearch",
+                "notexists",
+                "roomlistcollapsed",
+                "showaway",
+                "decimalsep"
+            ),
+            "columns": (
+                "downloads",
+                "uploads",
+                "search",
+                "search_widths",
+                "downloads_columns",
+                "downloads_widths",
+                "uploads_columns",
+                "uploads_widths",
+                "userbrowse",
+                "userbrowse_widths",
+                "userlist",
+                "userlist_widths",
+                "chatrooms",
+                "chatrooms_widths",
+                "download_columns",
+                "download_widths",
+                "upload_columns",
+                "upload_widths",
+                "filesearch_columns",
+                "filesearch_widths",
+                "hideflags"
+            ),
+            "searches": (
+                "distrib_timer",
+                "distrib_ignore",
+                "reopen_tabs",
+                "max_stored_results",
+                "re_filter"
+            ),
+            "userinfo": (
+                "descrutf8"
+            ),
+            "private_rooms": (
+                "enabled"
+            ),
+            "logging": (
+                "logsdir"
+            ),
+            "ticker": (
+                "default",
+                "rooms",
+                "hide"
+            ),
+            "language": (
+                "language",
+                "setlanguage"
+            ),
+            "urls": (
+                "urlcatching",
+                "humanizeurls"
+            ),
+            "notifications": (
+                "notification_tab_icons"
+            )
+        }
+
         # Windows specific stuff
         if sys.platform == "win32":
             self.defaults['players']['npplayer'] = 'other'
@@ -465,12 +555,22 @@ class Config:
 
         load_file(self.filename, self.parse_config)
 
-        # Clean up old config options
-        self.remove_old_options()
-
         # Update config values from file
         self.set_config()
 
+        if sys.platform == "darwin":
+            # Disable header bar in macOS for now due to GTK 3 performance issues
+            self.sections["ui"]["header_bar"] = False
+
+        # Convert special download folder share to regular share
+        if self.sections["transfers"].get("sharedownloaddir", False):
+            shares = self.sections["transfers"]["shared"]
+            virtual_name = "Downloaded"
+            shared_folder = (virtual_name, self.sections["transfers"]["downloaddir"])
+
+            if shared_folder not in shares and virtual_name not in (x[0] for x in shares):
+                shares.append(shared_folder)
+
         # Load command aliases from legacy file
         try:
             if not self.sections["server"]["command_aliases"] and os.path.exists(self.filename + ".alias"):
@@ -482,6 +582,8 @@ class Config:
         except Exception:
             return
 
+        self.config_loaded = True
+
     def parse_config(self, filename):
         """ Parses the config file """
 
@@ -502,7 +604,10 @@ class Config:
             from chardet import detect
 
         except ImportError:
-            print("Failed to convert config file to UTF-8. Please install python3-chardet and start Nicotine+ again.")
+            from pynicotine.logfacility import log
+
+            log.add("Failed to convert config file to UTF-8. Please install python3-chardet and start "
+                    "the application again.")
             sys.exit()
 
         os.rename(self.filename, self.filename + ".conv")
@@ -512,7 +617,7 @@ class Config:
 
         from_encoding = detect(rawdata)['encoding']
 
-        with open(self.filename + ".conv", 'r', encoding=from_encoding) as file_read:
+        with open(self.filename + ".conv", encoding=from_encoding) as file_read:
             with open(self.filename, 'w', encoding="utf-8") as file_write:
                 for line in file_read:
                     file_write.write(line[:-1] + '\r\n')
@@ -536,11 +641,12 @@ class Config:
             for j, val in self.parser.items(i, raw=True):
 
                 # Check if config section exists in defaults
-                if i not in self.defaults:
+                if i not in self.defaults and i not in self.removed_options:
                     log.add(_("Unknown config section '%s'"), i)
 
                 # Check if config option exists in defaults
-                elif j not in self.defaults[i] and i != "plugins" and j != "filter":
+                elif (j not in self.defaults.get(i, "") and j not in self.removed_options.get(i, "")
+                        and i != "plugins" and j != "filter"):
                     log.add(_("Unknown config option '%(option)s' in section '%(section)s'"),
                             {'option': j, 'section': i})
 
@@ -589,9 +695,9 @@ class Config:
                         self.sections[i][j] = default_val
 
                         log.add("Config error: Couldn't decode '%s' section '%s' value '%s', value has been reset", (
-                            (i[:120] + '..') if len(i) > 120 else i,
-                            (j[:120] + '..') if len(j) > 120 else j,
-                            (val[:120] + '..') if len(val) > 120 else val
+                            (i[:120] + '…') if len(i) > 120 else i,
+                            (j[:120] + '…') if len(j) > 120 else j,
+                            (val[:120] + '…') if len(val) > 120 else val
                         ))
 
         server = self.sections["server"]
@@ -613,127 +719,29 @@ class Config:
             # Setting the port range in numerical order
             server["portrange"] = (min(server["portrange"]), max(server["portrange"]))
 
-    def remove_old_options(self):
-
-        # Transition from 1.2.16 -> 1.4.0
-        # Do the cleanup early so we don't get the annoying
-        # 'Unknown config option ...' message
-        self.remove_old_option("transfers", "pmqueueddir")
-        self.remove_old_option("server", "lastportstatuscheck")
-        self.remove_old_option("server", "serverlist")
-        self.remove_old_option("userinfo", "descrutf8")
-        self.remove_old_option("ui", "enabletrans")
-        self.remove_old_option("ui", "mozembed")
-        self.remove_old_option("ui", "open_in_mozembed")
-        self.remove_old_option("ui", "tooltips")
-        self.remove_old_option("ui", "transalpha")
-        self.remove_old_option("ui", "transfilter")
-        self.remove_old_option("ui", "transtint")
-        self.remove_old_option("language", "language")
-        self.remove_old_option("language", "setlanguage")
-        self.remove_old_section("language")
-
-        # Transition from 1.4.1 -> 1.4.2
-        self.remove_old_option("columns", "downloads")
-        self.remove_old_option("columns", "uploads")
-
-        # Remove old encoding settings (1.4.3)
-        self.remove_old_option("server", "enc")
-        self.remove_old_option("server", "fallbackencodings")
-        self.remove_old_option("server", "roomencoding")
-        self.remove_old_option("server", "userencoding")
-
-        # Remove soundcommand config, replaced by GSound (1.4.3)
-        self.remove_old_option("ui", "soundcommand")
-
-        # Remove old column widths in preparation for "group by folder"-feature
-        self.remove_old_option("columns", "search")
-        self.remove_old_option("columns", "search_widths")
-        self.remove_old_option("columns", "downloads_columns")
-        self.remove_old_option("columns", "downloads_widths")
-        self.remove_old_option("columns", "uploads_columns")
-        self.remove_old_option("columns", "uploads_widths")
-
-        # Remove auto-retry failed downloads-option, this is now default behavior
-        self.remove_old_option("transfers", "autoretry_downloads")
-
-        # Remove old notification/sound settings
-        self.remove_old_option("transfers", "shownotification")
-        self.remove_old_option("transfers", "shownotificationperfolder")
-        self.remove_old_option("ui", "soundenabled")
-        self.remove_old_option("ui", "soundtheme")
-        self.remove_old_option("ui", "tab_colors")
-        self.remove_old_option("ui", "tab_icons")
-
-        # Remove dropped offline user text color in search results
-        self.remove_old_option("ui", "searchoffline")
-
-        # Seems to be superseded by ("server", "private_chatrooms")
-        self.remove_old_option("private_rooms", "enabled")
-
-        # Remove everything ticker-related, no longer necessary after the introduction of room walls
-        self.remove_old_section("ticker")
-
-        # Remove old log folder option, superseded by individual log folders for transfers and debug messages
-        self.remove_old_option("logging", "logsdir")
-
-        # Remove option to stop responding to searches for a certain time
-        self.remove_old_option("searches", "distrib_timer")
-        self.remove_old_option("searches", "distrib_ignore")
-
-        # Remove "I can receive direct connections"-option, it's redundant now
-        self.remove_old_option("server", "firewalled")
-
-        # Remove old column options
-        self.remove_old_option("columns", "userbrowse")
-        self.remove_old_option("columns", "userbrowse_widths")
-        self.remove_old_option("columns", "userlist")
-        self.remove_old_option("columns", "userlist_widths")
-        self.remove_old_option("columns", "chatrooms")
-        self.remove_old_option("columns", "chatrooms_widths")
-        self.remove_old_option("columns", "download_columns")
-        self.remove_old_option("columns", "download_widths")
-        self.remove_old_option("columns", "upload_columns")
-        self.remove_old_option("columns", "upload_widths")
-        self.remove_old_option("columns", "filesearch_columns")
-        self.remove_old_option("columns", "filesearch_widths")
-
-        # Remove option to prioritize sfv/md5 files when downloading
-        self.remove_old_option("transfers", "prioritize")
-
-        # Remove option to reopen closed search tabs when new results come in
-        self.remove_old_option("searches", "reopen_tabs")
-
-        # Remove option to toggle chatroom arrow buttons, since they no longer exist
-        self.remove_old_option("ui", "chat_hidebuttons")
-
-        # Remove max stored search results, only visible search result limit is used now
-        self.remove_old_option("searches", "max_stored_results")
-
-        # Remove option to specify if tabs are reorderable, since there's no clear advantage
-        self.remove_old_option("ui", "tab_reorderable")
-
-    def remove_old_option(self, section, option):
-
-        if section in self.parser.sections() and option in self.parser.options(section):
-            self.parser.remove_option(section, option)
-
-    def remove_old_section(self, section):
-
-        if section in self.parser.sections():
-            self.parser.remove_section(section)
-
     def write_config_callback(self, filename):
         self.parser.write(filename)
 
     def write_configuration(self):
 
-        for i in self.sections:
-            if not self.parser.has_section(i):
-                self.parser.add_section(i)
+        if not self.config_loaded:
+            return
 
-            for j in self.sections[i]:
-                self.parser.set(i, j, str(self.sections[i][j]))
+        # Write new config options to file
+        for section, options in self.sections.items():
+            if not self.parser.has_section(section):
+                self.parser.add_section(section)
+
+            for option, value in options.items():
+                self.parser.set(section, option, str(value))
+
+        # Remove legacy config options
+        for section, options in self.removed_options.items():
+            if not self.parser.has_section(section):
+                continue
+
+            for option in options:
+                self.parser.remove_option(section, option)
 
         if not self.create_config_folder():
             return
@@ -743,6 +751,8 @@ class Config:
 
     def write_config_backup(self, filename):
 
+        from pynicotine.logfacility import log
+
         if not filename.endswith(".tar.bz2"):
             filename += ".tar.bz2"
 
@@ -758,10 +768,10 @@ class Config:
                 tar.add(self.filename)
 
         except Exception as error:
-            print(error)
-            return (True, "Cannot write backup archive: %s" % error)
+            log.add(_("Error backing up config: %s"), error)
+            return
 
-        return (False, filename)
+        log.add(_("Config backed up to: %s"), filename)
 
 
 config = Config()
