@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2021 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Team
 # COPYRIGHT (C) 2016-2018 Mutnick <mutnick@techie.com>
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2009-2011 Quinox <quinox@users.sf.net>
@@ -26,50 +26,75 @@ import os
 
 from pynicotine.config import config
 from pynicotine.gtkgui.transferlist import TransferList
-from pynicotine.gtkgui.utils import open_file_path
+from pynicotine.gtkgui.utils import copy_text
 from pynicotine.gtkgui.widgets.dialogs import option_dialog
+from pynicotine.utils import open_file_path
 
 
 class Uploads(TransferList):
 
-    def __init__(self, frame, tab_label):
+    def __init__(self, frame):
 
-        TransferList.__init__(self, frame, type='upload')
-        self.tab_label = tab_label
+        self.path_separator = '\\'
+        self.path_label = _("Folder")
+        self.retry_label = _("_Retry")
+        self.abort_label = _("_Abort")
+        self.aborted_status = "Aborted"
 
-        self.popup_menu_clear.setup(
+        TransferList.__init__(self, frame, transfer_type="upload")
+
+        self.popup_menu_clear.add_items(
             ("#" + _("Finished / Aborted / Failed"), self.on_clear_finished_failed),
             ("#" + _("Finished / Aborted"), self.on_clear_finished_aborted),
             ("", None),
             ("#" + _("Finished"), self.on_clear_finished),
             ("#" + _("Aborted"), self.on_clear_aborted),
             ("#" + _("Failed"), self.on_clear_failed),
-            ("#" + _("Queued..."), self.on_try_clear_queued),
+            ("#" + _("Queued…"), self.on_try_clear_queued),
             ("", None),
-            ("#" + _("Clear All..."), self.on_try_clear_all),
+            ("#" + _("Clear All…"), self.on_try_clear_all),
         )
+        self.popup_menu_clear.update_model()
 
-    def on_try_clear_queued(self, *args):
+    def on_try_clear_queued(self, *_args):
 
         option_dialog(
             parent=self.frame.MainWindow,
             title=_('Clear Queued Uploads'),
-            message=_('Are you sure you wish to clear all queued uploads?'),
+            message=_('Do you really want to clear all queued uploads?'),
             callback=self.on_clear_response,
             callback_data="queued"
         )
 
-    def on_try_clear_all(self, *args):
+    def on_try_clear_all(self, *_args):
 
         option_dialog(
             parent=self.frame.MainWindow,
             title=_('Clear All Uploads'),
-            message=_('Are you sure you wish to clear all uploads?'),
+            message=_('Do you really want to clear all uploads?'),
             callback=self.on_clear_response,
             callback_data="all"
         )
 
-    def on_open_directory(self, *args):
+    def on_copy_url(self, *_args):
+
+        transfer = next(iter(self.selected_transfers), None)
+
+        if transfer:
+            user = config.sections["server"]["login"]
+            url = self.frame.np.userbrowse.get_soulseek_url(user, transfer.filename)
+            copy_text(url)
+
+    def on_copy_dir_url(self, *_args):
+
+        transfer = next(iter(self.selected_transfers), None)
+
+        if transfer:
+            user = config.sections["server"]["login"]
+            url = self.frame.np.userbrowse.get_soulseek_url(user, transfer.filename.rsplit('\\', 1)[0] + '\\')
+            copy_text(url)
+
+    def on_open_file_manager(self, *_args):
 
         transfer = next(iter(self.selected_transfers), None)
 
@@ -80,7 +105,7 @@ class Uploads(TransferList):
         command = config.sections["ui"]["filemanager"]
         open_file_path(transfer.path, command)
 
-    def on_play_files(self, *args):
+    def on_play_files(self, *_args):
 
         for transfer in self.selected_transfers:
             basename = str.split(transfer.filename, '\\')[-1]
@@ -90,7 +115,7 @@ class Uploads(TransferList):
                 command = config.sections["players"]["default"]
                 open_file_path(playfile, command)
 
-    def on_browse_folder(self, *args):
+    def on_browse_folder(self, *_args):
 
         transfer = next(iter(self.selected_transfers), None)
 
@@ -98,30 +123,30 @@ class Uploads(TransferList):
             return
 
         user = config.sections["server"]["login"]
-        folder = transfer.filename.rsplit('\\', 1)[0]
+        folder = transfer.filename.rsplit('\\', 1)[0] + '\\'
 
-        self.frame.np.userbrowse.browse_user(user, folder=folder)
+        self.frame.np.userbrowse.browse_user(user, path=folder)
 
-    def on_abort_user(self, *args):
+    def on_abort_user(self, *_args):
 
         self.select_transfers()
 
         for user in self.selected_users:
             for transfer in self.transfer_list:
-                if transfer.user == user:
-                    self.selected_transfers.add(transfer)
+                if transfer.user == user and transfer not in self.selected_transfers:
+                    self.selected_transfers.append(transfer)
 
         self.abort_transfers()
 
-    def on_clear_aborted(self, *args):
+    def on_clear_aborted(self, *_args):
         self.clear_transfers(["Aborted", "Cancelled", "Disallowed extension", "User logged off"])
 
-    def on_clear_failed(self, *args):
+    def on_clear_failed(self, *_args):
         self.clear_transfers(["Cannot connect", "Local file error", "Remote file error"])
 
-    def on_clear_finished_aborted(self, *args):
+    def on_clear_finished_aborted(self, *_args):
         self.clear_transfers(["Aborted", "Cancelled", "Disallowed extension", "User logged off", "Finished"])
 
-    def on_clear_finished_failed(self, *args):
+    def on_clear_finished_failed(self, *_args):
         self.clear_transfers(["Aborted", "Cancelled", "Disallowed extension", "User logged off", "Finished",
                               "Cannot connect", "Local file error", "Remote file error"])
