@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2021 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Team
 # COPYRIGHT (C) 2018 Mutnick <mutnick@techie.com>
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2009 Quinox <quinox@users.sf.net>
@@ -21,50 +21,51 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import os
 import time
 
-from gi.repository import GLib
+from gi.repository import Gio
 from gi.repository import GObject
 from gi.repository import Gtk
 
 from pynicotine.config import config
-from pynicotine.gtkgui.utils import load_ui_elements
 from pynicotine.gtkgui.widgets.dialogs import entry_dialog
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
+from pynicotine.gtkgui.widgets.textentry import CompletionEntry
+from pynicotine.gtkgui.widgets.theme import get_flag_icon_name
+from pynicotine.gtkgui.widgets.theme import get_status_icon
 from pynicotine.gtkgui.widgets.theme import update_widget_visuals
 from pynicotine.gtkgui.widgets.treeview import initialise_columns
 from pynicotine.gtkgui.widgets.treeview import save_columns
 from pynicotine.gtkgui.widgets.treeview import show_country_tooltip
 from pynicotine.gtkgui.widgets.treeview import show_user_status_tooltip
+from pynicotine.gtkgui.widgets.ui import UserInterface
 from pynicotine.logfacility import log
 from pynicotine.utils import humanize
 from pynicotine.utils import human_speed
 
 
-class UserList:
+class UserList(UserInterface):
 
     def __init__(self, frame):
 
-        # Build the window
+        super().__init__("ui/buddylist.ui")
+
         self.frame = frame
+        self.page_id = "userlist"
 
-        load_ui_elements(self, os.path.join(self.frame.gui_dir, "ui", "buddylist.ui"))
-
-        """ Columns """
-
+        # Columns
         self.user_iterators = {}
         self.usersmodel = Gtk.ListStore(
-            GObject.TYPE_OBJECT,  # (0)  status icon
-            GObject.TYPE_OBJECT,  # (1)  flag
+            Gio.Icon,             # (0)  status icon
+            str,                  # (1)  flag
             str,                  # (2)  username
             str,                  # (3)  hspeed
             str,                  # (4)  hfile count
             bool,                 # (5)  trusted
             bool,                 # (6)  notify
-            bool,                 # (7)  privileged
+            bool,                 # (7)  prioritized
             str,                  # (8)  hlast seen
-            str,                  # (9)  comments
+            str,                  # (9)  note
             int,                  # (10) status
             GObject.TYPE_UINT64,  # (11) speed
             GObject.TYPE_UINT64,  # (12) file count
@@ -74,17 +75,17 @@ class UserList:
 
         self.column_numbers = list(range(self.usersmodel.get_n_columns()))
         self.cols = cols = initialise_columns(
-            "buddy_list", self.UserListTree,
-            ["status", _("Status"), 25, "pixbuf", None],
-            ["country", _("Country"), 25, "pixbuf", None],
+            frame, "buddy_list", self.UserListTree,
+            ["status", _("Status"), 25, "icon", None],
+            ["country", _("Country"), 25, "icon", None],
             ["user", _("User"), 250, "text", None],
             ["speed", _("Speed"), 150, "number", None],
             ["files", _("Files"), 150, "number", None],
             ["trusted", _("Trusted"), 0, "toggle", None],
             ["notify", _("Notify"), 0, "toggle", None],
-            ["privileged", _("Privileged"), 0, "toggle", None],
-            ["last_seen", _("Last seen"), 160, "text", None],
-            ["comments", _("Comments"), 400, "edit", None]
+            ["privileged", _("Prioritized"), 0, "toggle", None],
+            ["last_seen", _("Last Seen"), 160, "text", None],
+            ["comments", _("Note"), 400, "edit", None]
         )
 
         cols["status"].set_sort_column_id(10)
@@ -101,9 +102,6 @@ class UserList:
         cols["status"].get_widget().hide()
         cols["country"].get_widget().hide()
 
-        if config.sections["columns"]["hideflags"]:
-            cols["country"].set_visible(False)
-
         for render in cols["trusted"].get_cells():
             render.connect('toggled', self.cell_toggle_callback, self.UserListTree, 5)
 
@@ -118,33 +116,28 @@ class UserList:
 
         self.UserListTree.set_model(self.usersmodel)
 
-        """ Lists """
-
+        # Lists
         for row in config.sections["server"]["userlist"]:
             self.append_user_row(row)
 
         self.usersmodel.set_sort_column_id(2, Gtk.SortType.ASCENDING)
 
-        self.buddies_combo_entries = (
-            self.frame.UserSearchCombo, self.frame.PrivateChatCombo, self.frame.UserInfoCombo,
-            self.frame.UserBrowseCombo)
+        for combo_box in (self.frame.UserSearchCombo, self.frame.PrivateChatCombo, self.frame.UserInfoCombo,
+                          self.frame.UserBrowseCombo):
+            combo_box.set_model(self.usersmodel)
+            combo_box.set_entry_text_column(2)
 
-        self.buddies_combos_fill()
-
-        """ Popup """
+            CompletionEntry(combo_box.get_child(), self.usersmodel, column=2)
 
+        # Popup menus
         self.popup_menu_private_rooms = PopupMenu(self.frame)
 
         self.popup_menu = popup = PopupMenu(frame, self.UserListTree, self.on_popup_menu)
         popup.setup_user_menu(page="userlist")
-        popup.setup(
-            ("", None),
-            ("$" + _("_Online Notify"), self.on_notify),
-            ("$" + _("_Privileged"), self.on_privileged),
-            ("$" + _("_Trusted"), self.on_trusted),
+        popup.add_items(
             ("", None),
+            ("#" + _("Add User _Note…"), self.on_add_note),
             (">" + _("Private Rooms"), self.popup_menu_private_rooms),
-            ("#" + _("Edit _Comments..."), self.on_edit_comments),
             ("#" + _("_Remove"), self.on_remove_user)
         )
 
@@ -161,9 +154,9 @@ class UserList:
             return
 
         try:
-            comment = str(row[1])
+            note = str(row[1])
         except IndexError:
-            comment = ""
+            note = ""
 
         try:
             notify = bool(row[2])
@@ -171,9 +164,9 @@ class UserList:
             notify = False
 
         try:
-            privileged = bool(row[3])
+            prioritized = bool(row[3])
         except IndexError:
-            privileged = False
+            prioritized = False
 
         try:
             trusted = bool(row[4])
@@ -197,16 +190,16 @@ class UserList:
             country = ""
 
         row = [
-            GObject.Value(GObject.TYPE_OBJECT, self.frame.get_status_image(0)),
-            GObject.Value(GObject.TYPE_OBJECT, self.frame.get_flag_image(country)),
+            get_status_icon(0),
+            get_flag_icon_name(country),
             username,
             "",
             "",
             trusted,
             notify,
-            privileged,
+            prioritized,
             last_seen,
-            comment,
+            note,
             0,
             0,
             0,
@@ -216,20 +209,11 @@ class UserList:
 
         self.user_iterators[username] = self.usersmodel.insert_with_valuesv(0, self.column_numbers, row)
 
-    def buddies_combos_fill(self):
+    @staticmethod
+    def on_tooltip(widget, pos_x, pos_y, _keyboard_mode, tooltip):
 
-        for widget in self.buddies_combo_entries:
-            widget.remove_all()
-            widget.append_text("")
-
-            for row in config.sections["server"]["userlist"]:
-                if row and isinstance(row, list):
-                    widget.append_text(str(row[0]))
-
-    def on_tooltip(self, widget, x, y, keyboard_mode, tooltip):
-
-        status_tooltip = show_user_status_tooltip(widget, x, y, tooltip, 10)
-        country_tooltip = show_country_tooltip(widget, x, y, tooltip, 14)
+        status_tooltip = show_user_status_tooltip(widget, pos_x, pos_y, tooltip, 10)
+        country_tooltip = show_country_tooltip(widget, pos_x, pos_y, tooltip, 14)
 
         if status_tooltip:
             return status_tooltip
@@ -237,7 +221,9 @@ class UserList:
         if country_tooltip:
             return country_tooltip
 
-    def on_add_user(self, widget, *args):
+        return False
+
+    def on_add_user(self, widget, *_args):
 
         username = widget.get_text()
 
@@ -247,15 +233,20 @@ class UserList:
         widget.set_text("")
         self.frame.np.userlist.add_user(username)
 
+    def update(self):
+
+        if config.sections["ui"]["buddylistinchatrooms"] in ("always", "chatrooms"):
+            return
+
+        self.frame.userlist_status_page.set_visible(not self.user_iterators)
+        self.Main.set_visible(self.user_iterators)
+
     def update_visuals(self):
 
         for widget in list(self.__dict__.values()):
             update_widget_visuals(widget)
 
-    def on_settings_ban_ignore(self, *args):
-        self.frame.on_settings_ban_ignore()
-
-    def cell_toggle_callback(self, widget, index, treeview, pos):
+    def cell_toggle_callback(self, _widget, index, treeview, pos):
 
         store = treeview.get_model()
         iterator = store.get_iter(index)
@@ -265,7 +256,7 @@ class UserList:
 
         self.save_user_list()
 
-    def cell_edited_callback(self, widget, index, value, treeview, pos):
+    def cell_edited_callback(self, _widget, index, value, treeview, pos):
 
         if pos != 9:
             return
@@ -273,7 +264,7 @@ class UserList:
         store = treeview.get_model()
         iterator = store.get_iter(index)
 
-        self.set_comment(iterator, store, value)
+        self.set_note(iterator, store, value)
 
     def set_last_seen(self, user, online=False):
 
@@ -295,13 +286,14 @@ class UserList:
         if not online:
             self.save_user_list()
 
-    def set_comment(self, iterator, store, comments=None):
+    def set_note(self, iterator, store, note=None):
 
-        if comments is not None:
-            store.set_value(iterator, 9, comments)
+        if note is not None:
+            store.set_value(iterator, 9, note)
             self.save_user_list()
 
-    def get_selected_username(self, treeview):
+    @staticmethod
+    def get_selected_username(treeview):
 
         model, iterator = treeview.get_selection().get_selected()
 
@@ -310,23 +302,21 @@ class UserList:
 
         return model.get_value(iterator, 2)
 
-    def get_selected_username_details(self, treeview):
+    @staticmethod
+    def get_selected_username_details(treeview):
 
         model, iterator = treeview.get_selection().get_selected()
 
         if iterator is not None:
             username = model.get_value(iterator, 2)
-            trusted = model.get_value(iterator, 5)
-            notify = model.get_value(iterator, 6)
-            privileged = model.get_value(iterator, 7)
             status = model.get_value(iterator, 10)
 
         else:
-            username = trusted = notify = privileged = status = None
+            username = status = None
 
-        return username, trusted, notify, privileged, status
+        return username, status
 
-    def on_row_activated(self, treeview, path, column):
+    def on_row_activated(self, treeview, _path, _column):
 
         user = self.get_selected_username(treeview)
 
@@ -336,23 +326,15 @@ class UserList:
 
     def on_popup_menu(self, menu, widget):
 
-        username, trusted, notify, privileged, status = self.get_selected_username_details(widget)
-        if username is None:
-            return True
-
+        username, status = self.get_selected_username_details(widget)
         menu.set_user(username)
         menu.toggle_user_items()
         menu.populate_private_rooms(self.popup_menu_private_rooms)
 
-        actions = menu.get_actions()
+        private_rooms_enabled = (self.popup_menu_private_rooms.items
+                                 and status > 0 and menu.user != self.frame.np.login_username)
 
-        actions[_("Private Rooms")].set_enabled(
-            status or menu.user != config.sections["server"]["login"]
-        )
-
-        actions[_("_Online Notify")].set_state(GLib.Variant.new_boolean(notify))
-        actions[_("_Privileged")].set_state(GLib.Variant.new_boolean(privileged))
-        actions[_("_Trusted")].set_state(GLib.Variant.new_boolean(trusted))
+        menu.actions[_("Private Rooms")].set_enabled(private_rooms_enabled)
 
     def get_user_status(self, msg):
 
@@ -384,9 +366,9 @@ class UserList:
             log.add(status_text, user)
             self.frame.notifications.new_text_notification(status_text % user)
 
-        img = self.frame.get_status_image(status)
-        self.usersmodel.set_value(iterator, 0, GObject.Value(GObject.TYPE_OBJECT, img))
-        self.usersmodel.set_value(iterator, 10, GObject.Value(GObject.TYPE_INT64, status))
+        status_icon = get_status_icon(status)
+        self.usersmodel.set_value(iterator, 0, status_icon)
+        self.usersmodel.set_value(iterator, 10, status)
 
         if status:  # online
             self.set_last_seen(user, online=True)
@@ -402,28 +384,34 @@ class UserList:
         if iterator is None:
             return
 
-        hspeed = human_speed(msg.avgspeed)
-        hfiles = humanize(msg.files)
+        h_speed = ""
+        avgspeed = msg.avgspeed
 
-        self.usersmodel.set_value(iterator, 3, hspeed)
-        self.usersmodel.set_value(iterator, 4, hfiles)
-        self.usersmodel.set_value(iterator, 11, GObject.Value(GObject.TYPE_UINT64, msg.avgspeed))
-        self.usersmodel.set_value(iterator, 12, GObject.Value(GObject.TYPE_UINT64, msg.files))
+        if avgspeed > 0:
+            h_speed = human_speed(avgspeed)
 
-    def set_user_country(self, user, country):
+        files = msg.files
+        h_files = humanize(files)
+
+        self.usersmodel.set_value(iterator, 3, h_speed)
+        self.usersmodel.set_value(iterator, 4, h_files)
+        self.usersmodel.set_value(iterator, 11, GObject.Value(GObject.TYPE_UINT64, avgspeed))
+        self.usersmodel.set_value(iterator, 12, GObject.Value(GObject.TYPE_UINT64, files))
+
+    def set_user_country(self, user, country_code):
 
         iterator = self.user_iterators.get(user)
 
         if iterator is None:
             return
 
-        flag_image = self.frame.get_flag_image(country)
+        flag_icon = get_flag_icon_name(country_code or "")
 
-        if not flag_image:
+        if not flag_icon:
             return
 
-        self.usersmodel.set_value(iterator, 1, GObject.Value(GObject.TYPE_OBJECT, flag_image))
-        self.usersmodel.set_value(iterator, 14, "flag_" + country)
+        self.usersmodel.set_value(iterator, 1, flag_icon)
+        self.usersmodel.set_value(iterator, 14, "flag_" + country_code)
 
     def add_user(self, user):
 
@@ -436,8 +424,8 @@ class UserList:
         self.user_iterators[user] = self.usersmodel.insert_with_valuesv(
             -1, self.column_numbers,
             [
-                GObject.Value(GObject.TYPE_OBJECT, self.frame.get_status_image(0)),
-                GObject.Value(GObject.TYPE_OBJECT, None),
+                get_status_icon(0),
+                empty_str,
                 user,
                 empty_str,
                 empty_str,
@@ -455,9 +443,7 @@ class UserList:
         )
 
         self.save_user_list()
-
-        for widget in self.buddies_combo_entries:
-            widget.append_text(user)
+        self.update()
 
         if config.sections["words"]["buddies"]:
             self.frame.update_completions()
@@ -469,7 +455,7 @@ class UserList:
             del self.user_iterators[user]
 
         self.save_user_list()
-        self.buddies_combos_fill()
+        self.update()
 
         if config.sections["words"]["buddies"]:
             self.frame.update_completions()
@@ -479,9 +465,9 @@ class UserList:
         user_list = []
 
         for i in self.usersmodel:
-            (status_icon, flag, user, hspeed, hfile_count, trusted, notify, privileged,
-                hlast_seen, comments, status, speed, file_count, last_seen, country) = i
-            user_list.append([user, comments, notify, privileged, trusted, hlast_seen, country])
+            (_status_icon, _flag, user, _hspeed, _hfile_count, trusted, notify, prioritized,
+                hlast_seen, note, _status, _speed, _file_count, _last_seen, country) = i
+            user_list.append([user, note, notify, prioritized, trusted, hlast_seen, country])
 
         self.frame.np.userlist.save_user_list(user_list)
 
@@ -514,7 +500,7 @@ class UserList:
         self.save_user_list()
         action.set_state(state)
 
-    def on_privileged(self, action, state):
+    def on_prioritized(self, action, state):
 
         user = self.popup_menu.get_user()
         iterator = self.user_iterators.get(user)
@@ -527,23 +513,23 @@ class UserList:
         self.save_user_list()
         action.set_state(state)
 
-    def on_edit_comments_response(self, dialog, response_id, user):
+    def on_add_note_response(self, dialog, _response_id, user):
 
         iterator = self.user_iterators.get(user)
 
         if iterator is None:
             return
 
-        comments = dialog.get_response_value()
+        note = dialog.get_response_value()
         dialog.destroy()
 
-        if comments is None:
+        if note is None:
             return
 
-        self.usersmodel.set_value(iterator, 9, comments)
+        self.usersmodel.set_value(iterator, 9, note)
         self.save_user_list()
 
-    def on_edit_comments(self, *args):
+    def on_add_note(self, *_args):
 
         user = self.popup_menu.get_user()
         iterator = self.user_iterators.get(user)
@@ -551,18 +537,18 @@ class UserList:
         if iterator is None:
             return
 
-        comments = self.usersmodel.get_value(iterator, 9) or ""
+        note = self.usersmodel.get_value(iterator, 9) or ""
 
         entry_dialog(
             parent=self.frame.MainWindow,
-            title=_("Edit Comments"),
-            message=_("Add a few notes associated with user %s:") % user,
-            callback=self.on_edit_comments_response,
+            title=_("Add User Note"),
+            message=_("Add a note about user %s:") % user,
+            callback=self.on_add_note_response,
             callback_data=user,
-            default=comments
+            default=note
         )
 
-    def on_remove_user(self, *args):
+    def on_remove_user(self, *_args):
         self.frame.np.userlist.remove_user(self.popup_menu.get_user())
 
     def server_disconnect(self):
@@ -570,7 +556,7 @@ class UserList:
         for i in self.usersmodel:
             iterator = i.iter
 
-            self.usersmodel.set_value(iterator, 0, GObject.Value(GObject.TYPE_OBJECT, self.frame.get_status_image(0)))
+            self.usersmodel.set_value(iterator, 0, get_status_icon(0))
             self.usersmodel.set_value(iterator, 3, "")
             self.usersmodel.set_value(iterator, 4, "")
             self.usersmodel.set_value(iterator, 10, 0)
