@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2021 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Team
 #
 # GNU GENERAL PUBLIC LICENSE
 #    Version 3, 29 June 2007
@@ -16,16 +16,16 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import gi
 import os
 import sys
+import gi
 
-from gi.repository import Gdk
-from gi.repository import GLib
 from gi.repository import Gtk
 
 from pynicotine.config import config
 from pynicotine.gtkgui.widgets.dialogs import entry_dialog
+from pynicotine.gtkgui.widgets.theme import get_icon
+from pynicotine.gtkgui.widgets.ui import GUI_DIR
 
 
 """ Status Icon / AppIndicator """
@@ -81,147 +81,116 @@ class TrayIcon:
             return
 
         self.tray_popup_menu = Gtk.Menu()
-        self.hide_show_item, handler = self.create_item(_("Show Nicotine+"), self.on_hide_unhide_window)
+        self.hide_show_item, _handler = self.create_item(_("Show Nicotine+"), self.frame.on_window_hide_unhide)
         self.alt_speed_item, self.alt_speed_handler = self.create_item(
             _("Alternative Speed Limits"), self.frame.on_alternative_speed_limit, check=True)
 
         self.tray_popup_menu.append(Gtk.SeparatorMenuItem())
 
-        self.downloads_item, handler = self.create_item(_("Downloads"), self.on_downloads)
-        self.uploads_item, handler = self.create_item(_("Uploads"), self.on_uploads)
+        self.downloads_item, _handler = self.create_item(_("Downloads"), self.on_downloads)
+        self.uploads_item, _handler = self.create_item(_("Uploads"), self.on_uploads)
 
         self.tray_popup_menu.append(Gtk.SeparatorMenuItem())
 
-        self.connect_item, handler = self.create_item(_("Connect"), self.frame.on_connect)
-        self.disconnect_item, handler = self.create_item(_("Disconnect"), self.frame.on_disconnect)
+        self.connect_item, _handler = self.create_item(_("Connect"), self.frame.on_connect)
+        self.disconnect_item, _handler = self.create_item(_("Disconnect"), self.frame.on_disconnect)
         self.away_item, self.away_handler = self.create_item(_("Away"), self.frame.on_away, check=True)
 
         self.tray_popup_menu.append(Gtk.SeparatorMenuItem())
 
-        self.send_message_item, handler = self.create_item(_("Send Message"), self.on_open_private_chat)
-        self.lookup_ip_item, handler = self.create_item(_("Receive a User's IP Address"), self.on_get_a_users_ip)
-        self.lookup_info_item, handler = self.create_item(_("Receive a User's Info"), self.on_get_a_users_info)
-        self.lookup_shares_item, handler = self.create_item(_("Receive a User's Shares"), self.on_get_a_users_shares)
+        self.send_message_item, _handler = self.create_item(_("Send Message"), self.on_open_private_chat)
+        self.lookup_info_item, _handler = self.create_item(_("Request User's Info"), self.on_get_a_users_info)
+        self.lookup_shares_item, _handler = self.create_item(_("Request User's Shares"), self.on_get_a_users_shares)
 
         self.tray_popup_menu.append(Gtk.SeparatorMenuItem())
 
         self.create_item(_("Preferences"), self.frame.on_settings)
-        self.create_item(_("Quit"), self.frame.on_quit)
+        self.create_item(_("Quit"), self.frame.np.quit)
 
-    def on_hide_unhide_window(self, *args):
+    def on_downloads(self, *_args):
+        self.frame.change_main_page("downloads")
+        self.frame.show()
 
-        if self.frame.MainWindow.get_property("visible"):
-            self.frame.MainWindow.hide()
-            return
+    def on_uploads(self, *_args):
+        self.frame.change_main_page("uploads")
+        self.frame.show()
 
-        self.show_window()
-
-    def on_downloads(self, *args):
-        self.frame.on_downloads()
-        self.show_window()
-
-    def on_uploads(self, *args):
-        self.frame.on_uploads()
-        self.show_window()
-
-    def on_open_private_chat_response(self, dialog, response_id, data):
+    def on_open_private_chat_response(self, dialog, response_id, _data):
 
         user = dialog.get_response_value()
         dialog.destroy()
 
-        if response_id != Gtk.ResponseType.OK:
+        if response_id != Gtk.ResponseType.OK or not user:
             return
 
-        if user:
-            self.frame.np.privatechats.show_user(user)
-            self.frame.change_main_page("private")
-            self.show_window()
+        self.frame.np.privatechats.show_user(user)
+        self.frame.change_main_page("private")
+        self.frame.show()
 
-    def on_open_private_chat(self, *args):
+    def on_open_private_chat(self, *_args):
 
         users = (i[0] for i in config.sections["server"]["userlist"])
         entry_dialog(
             parent=self.frame.application.get_active_window(),
-            title=GLib.get_application_name() + ": " + _("Start Messaging"),
-            message=_('Enter the name of a user who you wish to send a message:'),
+            title=config.application_name + ": " + _("Start Messaging"),
+            message=_('Enter the name of the user whom you want to send a message:'),
             callback=self.on_open_private_chat_response,
             droplist=users
         )
 
-    def on_get_a_users_info_response(self, dialog, response_id, data):
+    def on_get_a_users_info_response(self, dialog, response_id, _data):
 
         user = dialog.get_response_value()
         dialog.destroy()
 
-        if response_id != Gtk.ResponseType.OK:
+        if response_id != Gtk.ResponseType.OK or not user:
             return
 
-        if user:
-            self.frame.np.userinfo.request_user_info(user)
+        self.frame.np.userinfo.request_user_info(user)
+        self.frame.show()
 
-    def on_get_a_users_info(self, *args):
+    def on_get_a_users_info(self, *_args):
 
         users = (i[0] for i in config.sections["server"]["userlist"])
         entry_dialog(
             parent=self.frame.application.get_active_window(),
-            title=GLib.get_application_name() + ": " + _("Request User Info"),
-            message=_('Enter the name of a user whose info you wish to receive:'),
+            title=config.application_name + ": " + _("Request User Info"),
+            message=_('Enter the name of the user whose info you want to see:'),
             callback=self.on_get_a_users_info_response,
             droplist=users
         )
 
-    def on_get_a_users_ip_response(self, dialog, response_id, data):
+    def on_get_a_users_shares_response(self, dialog, response_id, _data):
 
         user = dialog.get_response_value()
         dialog.destroy()
 
-        if response_id != Gtk.ResponseType.OK:
+        if response_id != Gtk.ResponseType.OK or not user:
             return
 
-        if user:
-            self.frame.np.request_ip_address(user)
+        self.frame.np.userbrowse.browse_user(user)
+        self.frame.show()
 
-    def on_get_a_users_ip(self, *args):
+    def on_get_a_users_shares(self, *_args):
 
         users = (i[0] for i in config.sections["server"]["userlist"])
         entry_dialog(
             parent=self.frame.application.get_active_window(),
-            title=GLib.get_application_name() + ": " + _("Request IP Address"),
-            message=_('Enter the name of a user whose IP address you wish to receive:'),
-            callback=self.on_get_a_users_ip_response,
-            droplist=users
-        )
-
-    def on_get_a_users_shares_response(self, dialog, response_id, data):
-
-        user = dialog.get_response_value()
-        dialog.destroy()
-
-        if response_id != Gtk.ResponseType.OK:
-            return
-
-        if user:
-            self.frame.np.userbrowse.browse_user(user)
-
-    def on_get_a_users_shares(self, *args):
-
-        users = (i[0] for i in config.sections["server"]["userlist"])
-        entry_dialog(
-            parent=self.frame.application.get_active_window(),
-            title=GLib.get_application_name() + ": " + _("Request Shares List"),
-            message=_('Enter the name of a user whose shares list you wish to receive:'),
+            title=config.application_name + ": " + _("Request Shares List"),
+            message=_('Enter the name of the user whose shares you want to see:'),
             callback=self.on_get_a_users_shares_response,
             droplist=users
         )
 
     # GtkStatusIcon fallback
-    def on_status_icon_popup(self, status_icon, button, activate_time):
+    def on_status_icon_popup(self, _status_icon, button, _activate_time):
 
         if button == 3:
             time = Gtk.get_current_event_time()
             self.tray_popup_menu.popup(None, None, None, None, button, time)
 
-    def check_icon_path(self, icon_name, icon_path, icon_type="local"):
+    @staticmethod
+    def check_icon_path(icon_name, icon_path, icon_type="local"):
 
         """
         Check if tray icons exist in the specified icon path.
@@ -231,7 +200,7 @@ class TrayIcon:
         """
 
         if icon_type == "local":
-            icon_scheme = GLib.get_prgname() + "-" + icon_name + "."
+            icon_scheme = config.application_id + "-" + icon_name + "."
         else:
             icon_scheme = "trayicon_" + icon_name + "."
 
@@ -265,7 +234,7 @@ class TrayIcon:
             local_icon_path = os.path.join(sys.prefix, "share", "icons", "hicolor", "scalable", "apps")
         else:
             # Git folder
-            local_icon_path = os.path.join(self.frame.gui_dir, "icons", "hicolor", "scalable", "apps")
+            local_icon_path = os.path.join(GUI_DIR, "icons", "hicolor", "scalable", "apps")
 
         for icon_name in ("away", "connect", "disconnect", "msg"):
 
@@ -276,7 +245,6 @@ class TrayIcon:
 
             # Check if local icons exist
             if self.check_icon_path(icon_name, local_icon_path, icon_type="local"):
-                self.local_icons = True
                 return local_icon_path
 
         return None
@@ -304,9 +272,9 @@ class TrayIcon:
         if self.tray_icon is None:
             if self.appindicator is not None:
                 tray_icon = self.appindicator.Indicator.new(
-                    GLib.get_application_name(),
-                    "",
-                    self.appindicator.IndicatorCategory.APPLICATION_STATUS)
+                    id=config.application_name,
+                    icon_name="",
+                    category=self.appindicator.IndicatorCategory.APPLICATION_STATUS)
                 tray_icon.set_menu(self.tray_popup_menu)
 
                 # Action to hide/unhide main window when middle clicking the tray icon
@@ -320,21 +288,23 @@ class TrayIcon:
 
             else:
                 # GtkStatusIcon fallback
-                tray_icon = Gtk.StatusIcon()
-                tray_icon.set_tooltip_text(GLib.get_application_name())
-                tray_icon.connect("activate", self.on_hide_unhide_window)
+                tray_icon = Gtk.StatusIcon(tooltip_text=config.application_name)
+                tray_icon.connect("activate", self.frame.on_window_hide_unhide)
                 tray_icon.connect("popup-menu", self.on_status_icon_popup)
 
             self.tray_icon = tray_icon
+            self.set_server_actions_sensitive(False)
 
         self.set_alternative_speed_limit(config.sections["transfers"]["usealtlimits"])
 
         if use_trayicon or config.sections["ui"]["trayicon"]:
             self.show()
-            self.set_image(self.tray_status["status"])
+
+            # Gtk.StatusIcon.is_embedded() may not be true yet (observed in LXDE), force an icon update
+            self.set_icon(self.tray_status["status"], force_update=True)
             return
 
-        self.set_image("msg")
+        self.set_icon("msg", force_update=True)
         self.hide()
 
     def show(self):
@@ -384,9 +354,9 @@ class TrayIcon:
 
         self.hide_show_item.set_label(text)
 
-    def set_image(self, status=None):
+    def set_icon(self, status=None, force_update=False):
 
-        if not self.is_visible():
+        if not force_update and not self.is_visible():
             return
 
         if status is not None:
@@ -407,18 +377,15 @@ class TrayIcon:
         if self.custom_icons:
             icon_name = "trayicon_" + icon_name
         else:
-            icon_name = GLib.get_prgname() + "-" + icon_name
+            icon_name = config.application_id + "-" + icon_name
 
         if self.appindicator is not None:
-            self.tray_icon.set_icon_full(icon_name, GLib.get_application_name())
+            self.tray_icon.set_icon_full(icon_name, config.application_name)
 
         else:
             # GtkStatusIcon fallback
             if self.custom_icons:
-                self.tray_icon.set_from_pixbuf(
-                    self.frame.images[icon_name]
-                )
-
+                self.tray_icon.set_from_gicon(get_icon(icon_name))
             else:
                 self.tray_icon.set_from_icon_name(icon_name)
 
@@ -432,7 +399,7 @@ class TrayIcon:
         else:
             self.tray_status["status"] = "connect"
 
-        self.set_image()
+        self.set_icon()
 
         with self.away_item.handler_block(self.away_handler):
             # Temporarily disable handler, we only want to change the visual checkbox appearance
@@ -448,14 +415,14 @@ class TrayIcon:
         else:
             self.tray_status["status"] = "disconnect"
 
-        self.set_image()
+        self.set_icon()
 
     def set_server_actions_sensitive(self, status):
 
         if self.tray_icon is None:
             return
 
-        for i in (self.disconnect_item, self.away_item, self.send_message_item, self.lookup_ip_item,
+        for i in (self.disconnect_item, self.away_item, self.send_message_item,
                   self.lookup_info_item, self.lookup_shares_item):
 
             """ Disable menu items when disconnected from server """
@@ -463,13 +430,19 @@ class TrayIcon:
 
         self.connect_item.set_sensitive(not status)
 
-    def set_transfer_status(self, download, upload):
+    def set_download_status(self, status):
 
         if self.tray_icon is None:
             return
 
-        self.downloads_item.set_label(download)
-        self.uploads_item.set_label(upload)
+        self.downloads_item.set_label(status)
+
+    def set_upload_status(self, status):
+
+        if self.tray_icon is None:
+            return
+
+        self.uploads_item.set_label(status)
 
     def set_alternative_speed_limit(self, enable):
 
@@ -479,7 +452,3 @@ class TrayIcon:
         with self.alt_speed_item.handler_block(self.alt_speed_handler):
             # Temporarily disable handler, we only want to change the visual checkbox appearance
             self.alt_speed_item.set_active(enable)
-
-    def show_window(self):
-        self.frame.MainWindow.present_with_time(Gdk.CURRENT_TIME)
-        self.frame.MainWindow.deiconify()
