@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2021 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Team
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2008-2009 Quinox <quinox@users.sf.net>
 # COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
@@ -29,9 +29,12 @@ from collections import OrderedDict
 from gi.repository import Gio
 from gi.repository import GLib
 from gi.repository import Gtk
+from gi.repository import Pango
 
 from pynicotine.config import config
 from pynicotine.geoip.geoip import GeoIP
+from pynicotine.gtkgui.utils import copy_text
+from pynicotine.gtkgui.widgets.accelerator import Accelerator
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
 
 
@@ -60,7 +63,7 @@ def verify_grouping_mode(mode):
 def create_grouping_menu(window, active_mode, callback):
 
     action_id = "grouping" + ''.join(random.choice(string.digits) for _ in range(8))
-    menu = Gio.Menu.new()
+    menu = Gio.Menu()
 
     menuitem = Gio.MenuItem.new(_("Ungrouped"), "win." + action_id + "::ungrouped")
     menu.append_item(menuitem)
@@ -71,8 +74,8 @@ def create_grouping_menu(window, active_mode, callback):
     menuitem = Gio.MenuItem.new(_("Group by User"), "win." + action_id + "::user_grouping")
     menu.append_item(menuitem)
 
-    state = GLib.Variant.new_string(verify_grouping_mode(active_mode))
-    action = Gio.SimpleAction.new_stateful(action_id, GLib.VariantType.new("s"), state)
+    state = GLib.Variant("s", verify_grouping_mode(active_mode))
+    action = Gio.SimpleAction(name=action_id, parameter_type=GLib.VariantType("s"), state=state)
     action.connect("change-state", callback)
 
     window.add_action(action)
@@ -82,16 +85,15 @@ def create_grouping_menu(window, active_mode, callback):
 
 
 def select_user_row_iter(fmodel, sel, user_index, selected_user, iterator):
+
     while iterator is not None:
         user = fmodel.get_value(iterator, user_index)
 
         if selected_user == user:
-            sel.select_path(fmodel.get_path(iterator),)
+            sel.select_path(fmodel.get_path(iterator))
 
         child = fmodel.iter_children(iterator)
-
         select_user_row_iter(fmodel, sel, user_index, selected_user, child)
-
         iterator = fmodel.iter_next(iterator)
 
 
@@ -110,13 +112,24 @@ def collapse_treeview(treeview, grouping_mode):
             iterator = model.iter_next(iterator)
 
 
-def initialise_columns(treeview_name, treeview, *args):
+def initialise_columns(frame, treeview_name, treeview, *args):
 
     i = 0
     cols = OrderedDict()
+    num_cols = len(args)
     column_config = None
 
-    for (id, title, width, type, extra) in args:
+    # GTK 4 rows need more padding to match GTK 3
+    if Gtk.get_major_version() == 4:
+        progress_padding = 1
+        height_padding = 5
+    else:
+        progress_padding = 0
+        height_padding = 3
+
+    width_padding = 10
+
+    for column_id, title, width, column_type, extra in args:
         if treeview_name:
             try:
                 column_config = config.sections["columns"][treeview_name[0]][treeview_name[1]]
@@ -124,7 +137,7 @@ def initialise_columns(treeview_name, treeview, *args):
                 column_config = config.sections["columns"][treeview_name]
 
             try:
-                width = column_config[id]["width"]
+                width = column_config[column_id]["width"]
             except Exception:
                 # Invalid value
                 pass
@@ -132,56 +145,45 @@ def initialise_columns(treeview_name, treeview, *args):
         if not isinstance(width, int):
             width = 0
 
-        if Gtk.get_major_version() == 4:
-            # GTK 4 rows need more padding to match GTK 3
-            height_padding = 5
-        else:
-            height_padding = 3
+        xalign = 0
 
-        if type == "text":
-            renderer = Gtk.CellRendererText()
-            renderer.set_padding(10, height_padding)
+        if column_type == "text":
+            renderer = Gtk.CellRendererText(ellipsize=Pango.EllipsizeMode.END, xpad=width_padding, ypad=height_padding)
+            column = Gtk.TreeViewColumn(column_id, renderer, text=i)
 
-            column = Gtk.TreeViewColumn(id, renderer, text=i)
+        elif column_type == "number":
+            xalign = 1
+            renderer = Gtk.CellRendererText(xalign=xalign, xpad=width_padding, ypad=height_padding)
+            column = Gtk.TreeViewColumn(column_id, renderer, text=i)
+            column.set_alignment(xalign)
 
-        elif type == "center":
-            renderer = Gtk.CellRendererText()
-            renderer.set_property("xalign", 0.5)
+        elif column_type == "edit":
+            renderer = Gtk.CellRendererText(editable=True, xpad=width_padding, ypad=height_padding)
+            column = Gtk.TreeViewColumn(column_id, renderer, text=i)
 
-            column = Gtk.TreeViewColumn(id, renderer, text=i)
+        elif column_type == "progress":
+            renderer = Gtk.CellRendererProgress(ypad=progress_padding)
+            column = Gtk.TreeViewColumn(column_id, renderer, value=i)
 
-        elif type == "number":
-            renderer = Gtk.CellRendererText()
-            renderer.set_property("xalign", 0.9)
+        elif column_type == "toggle":
+            xalign = 0.5
+            renderer = Gtk.CellRendererToggle(xalign=xalign, xpad=13)
+            column = Gtk.TreeViewColumn(column_id, renderer, active=i)
 
-            column = Gtk.TreeViewColumn(id, renderer, text=i)
-            column.set_alignment(0.9)
-
-        elif type == "edit":
-            renderer = Gtk.CellRendererText()
-            renderer.set_padding(10, height_padding)
-            renderer.set_property('editable', True)
-            column = Gtk.TreeViewColumn(id, renderer, text=i)
-
-        elif type == "combo":
-            renderer = Gtk.CellRendererCombo()
-            renderer.set_padding(10, height_padding)
-            renderer.set_property('text-column', 0)
-            renderer.set_property('editable', True)
-            column = Gtk.TreeViewColumn(id, renderer, text=i)
-
-        elif type == "progress":
-            renderer = Gtk.CellRendererProgress()
-            column = Gtk.TreeViewColumn(id, renderer, value=i)
-
-        elif type == "toggle":
-            renderer = Gtk.CellRendererToggle()
-            column = Gtk.TreeViewColumn(id, renderer, active=i)
-            renderer.set_property("xalign", 0.5)
-
-        else:
+        elif column_type == "icon":
             renderer = Gtk.CellRendererPixbuf()
-            column = Gtk.TreeViewColumn(id, renderer, pixbuf=i)
+
+            if column_id == "country":
+                if Gtk.get_major_version() == 4:
+                    # Custom icon size defined in theme.py
+                    renderer.set_property("icon-size", Gtk.IconSize.NORMAL)
+                else:
+                    # Use the same size as the original icon
+                    renderer.set_property("stock-size", 0)
+
+                column = Gtk.TreeViewColumn(column_id, renderer, icon_name=i)
+            else:
+                column = Gtk.TreeViewColumn(column_id, renderer, gicon=i)
 
         if width == -1:
             column.set_resizable(False)
@@ -190,12 +192,13 @@ def initialise_columns(treeview_name, treeview, *args):
 
         else:
             column.set_resizable(True)
+            column.set_min_width(0)
+
             if width == 0:
                 column.set_sizing(Gtk.TreeViewColumnSizing.GROW_ONLY)
             else:
                 column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
                 column.set_fixed_width(width)
-            column.set_min_width(0)
 
         if isinstance(extra, int):
             column.add_attribute(renderer, "foreground", extra)
@@ -205,26 +208,22 @@ def initialise_columns(treeview_name, treeview, *args):
             column.add_attribute(renderer, "weight", weight)
             column.add_attribute(renderer, "underline", underline)
 
+        # Allow individual cells to receive visual focus
+        if num_cols > 1:
+            renderer.set_property("mode", Gtk.CellRendererMode.ACTIVATABLE)
+
         column.set_reorderable(True)
         column.set_min_width(20)
 
-        column.set_widget(Gtk.Label.new(title))
-        column.get_widget().set_margin_start(5)
-        column.get_widget().show()
+        label = Gtk.Label(label=title, margin_start=5, margin_end=5, visible=True)
+        column.set_widget(label)
 
-        if Gtk.get_major_version() == 4:
-            """ Temporary hack to restore sorting by clicking column headers in GTK 4 """
+        if xalign == 1 and Gtk.get_major_version() == 4:
+            # Gtk.TreeViewColumn.set_alignment() only changes the sort arrow position in GTK 4
+            # Actually align the label to the right here instead
+            label.get_parent().set_halign(Gtk.Align.END)
 
-            button = column.get_button()
-            button.set_overflow(Gtk.Overflow.HIDDEN)
-
-            gesture_click = Gtk.GestureClick()
-            gesture_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
-            gesture_click.connect(
-                "pressed", lambda controller, *args: controller.set_state(Gtk.EventSequenceState.CLAIMED))
-            button.add_controller(gesture_click)
-
-        cols[id] = column
+        cols[column_id] = column
 
         i += 1
 
@@ -234,19 +233,43 @@ def initialise_columns(treeview_name, treeview, *args):
     treeview.connect("columns-changed", set_last_column_autosize)
     treeview.emit("columns-changed")
 
+    Accelerator("<Primary>c", treeview, on_copy_cell_data_accelerator)
+    treeview.column_menu = PopupMenu(frame, treeview, callback=press_header, connect_events=False)
+
+    if Gtk.get_major_version() == 4:
+        # Hotfix: disable rubber-band selection in GTK 4 to avoid crash bug
+        # when clicking column headers
+        treeview.set_rubber_banding(False)
+
     return cols
 
 
-def append_columns(treeview, cols, config):
+def on_copy_cell_data_accelerator(treeview, *_args):
+    """ Ctrl+C: copy cell data """
+
+    path, column = treeview.get_cursor()
+    model = treeview.get_model()
+
+    if path is None:
+        return False
+
+    iterator = model.get_iter(path)
+    cell_value = str(model.get_value(iterator, column.get_sort_column_id()))
+
+    copy_text(cell_value)
+    return True
+
+
+def append_columns(treeview, cols, column_config):
 
     # Column order not supported in Python 3.5
-    if not config or sys.version_info[:2] <= (3, 5):
-        for (column_id, column) in cols.items():
+    if not column_config or sys.version_info[:2] <= (3, 5):
+        for column_id, column in cols.items():
             treeview.append_column(column)
         return
 
     # Restore column order from config
-    for column_id in config:
+    for column_id in column_config:
         try:
             treeview.append_column(cols[column_id])
         except Exception:
@@ -257,7 +280,7 @@ def append_columns(treeview, cols, config):
 
     # If any columns were missing in the config, append them
     pos = 0
-    for (column_id, column) in cols.items():
+    for column_id, column in cols.items():
         if column not in added_columns:
             treeview.insert_column(column, pos)
 
@@ -295,20 +318,18 @@ def set_last_column_autosize(treeview):
     prev_column.set_resizable(True)
 
 
-def hide_columns(treeview, cols, config):
-
-    for (column_id, column) in cols.items():
-        parent = column.get_button()
-        if parent:
-            PopupMenu(widget=parent, callback=press_header)
+def hide_columns(_treeview, cols, column_config):
 
+    for column_id, column in cols.items():
         # Read Show / Hide column settings from last session
-        if config:
-            try:
-                column.set_visible(config[column_id]["visible"])
-            except Exception:
-                # Invalid value
-                pass
+        if not column_config:
+            continue
+
+        try:
+            column.set_visible(column_config[column_id]["visible"])
+        except Exception:
+            # Invalid value
+            pass
 
 
 def save_columns(treeview_name, columns, subpage=None):
@@ -353,13 +374,16 @@ def save_columns(treeview_name, columns, subpage=None):
         column_config[treeview_name] = saved_columns
 
 
-def press_header(menu, widget):
+def press_header(menu, treeview):
 
-    treeview = widget.get_parent()
     columns = treeview.get_columns()
+
+    if len(columns) <= 1:
+        # Only a single column, don't show menu
+        return True
+
     visible_columns = [column for column in columns if column.get_visible()]
     menu.clear()
-    actions = menu.get_actions()
     pos = 1
 
     for column in columns:
@@ -368,22 +392,22 @@ def press_header(menu, widget):
         if title == "":
             title = _("Column #%i") % pos
 
-        menu.setup(
+        menu.add_items(
             ("$" + title, None)
         )
-
-        actions[title].set_state(
-            GLib.Variant.new_boolean(column in visible_columns)
-        )
+        menu.update_model()
+        menu.actions[title].set_state(GLib.Variant("b", column in visible_columns))
 
         if column in visible_columns:
-            actions[title].set_enabled(len(visible_columns) > 1)
+            menu.actions[title].set_enabled(len(visible_columns) > 1)
 
-        actions[title].connect("activate", header_toggle, treeview, columns, pos - 1)
+        menu.actions[title].connect("activate", header_toggle, treeview, columns, pos - 1)
         pos += 1
 
+    return False
 
-def header_toggle(action, state, treeview, columns, index):
+
+def header_toggle(_action, _state, treeview, columns, index):
 
     column = columns[index]
     column.set_visible(not column.get_visible())
@@ -397,22 +421,22 @@ def set_treeview_selected_row(treeview, bin_x, bin_y):
     selection = treeview.get_selection()
 
     if pathinfo is not None:
-        path, col, cell_x, cell_y = pathinfo
+        path, column, _cell_x, _cell_y = pathinfo
 
         # Make sure we don't attempt to select a single row if the row is already
         # in a selection of multiple rows, otherwise the other rows will be unselected
         if selection.count_selected_rows() <= 1 or not selection.path_is_selected(path):
             treeview.grab_focus()
-            treeview.set_cursor(path, col, False)
+            treeview.set_cursor(path, column, False)
     else:
         selection.unselect_all()
 
 
-def show_tooltip(treeview, x, y, tooltip, sourcecolumn, column_titles, text_function, strip_prefix=""):
+def show_tooltip(treeview, pos_x, pos_y, tooltip, sourcecolumn, column_titles, text_function, strip_prefix=""):
 
     try:
-        bin_x, bin_y = treeview.convert_widget_to_bin_window_coords(x, y)
-        path, column, cx, cy = treeview.get_path_at_pos(bin_x, bin_y)
+        bin_x, bin_y = treeview.convert_widget_to_bin_window_coords(pos_x, pos_y)
+        path, column, _cell_x, _cell_y = treeview.get_path_at_pos(bin_x, bin_y)
 
     except TypeError:
         return False
@@ -447,11 +471,15 @@ def get_country_tooltip_text(column_value, strip_prefix):
     return _("Earth")
 
 
-def get_file_path_tooltip_text(column_value, strip_prefix):
+def get_file_path_tooltip_text(column_value, _strip_prefix):
     return column_value
 
 
-def get_user_status_tooltip_text(column_value, strip_prefix):
+def get_transfer_file_path_tooltip_text(column_value, _strip_prefix):
+    return column_value.filename or column_value.path
+
+
+def get_user_status_tooltip_text(column_value, _strip_prefix):
 
     if column_value == 1:
         return _("Away")
@@ -462,17 +490,21 @@ def get_user_status_tooltip_text(column_value, strip_prefix):
     return _("Offline")
 
 
-def show_country_tooltip(treeview, x, y, tooltip, sourcecolumn, strip_prefix='flag_'):
-    return show_tooltip(treeview, x, y, tooltip, sourcecolumn, ("country",), get_country_tooltip_text, strip_prefix)
+def show_country_tooltip(treeview, pos_x, pos_y, tooltip, sourcecolumn, strip_prefix='flag_'):
+    return show_tooltip(treeview, pos_x, pos_y, tooltip, sourcecolumn,
+                        ("country",), get_country_tooltip_text, strip_prefix)
 
 
-def show_file_path_tooltip(treeview, x, y, tooltip, sourcecolumn):
+def show_file_path_tooltip(treeview, pos_x, pos_y, tooltip, sourcecolumn, transfer=False):
 
     if not config.sections["ui"]["file_path_tooltips"]:
         return False
 
-    return show_tooltip(treeview, x, y, tooltip, sourcecolumn, ("folder", "filename"), get_file_path_tooltip_text)
+    function = get_file_path_tooltip_text if not transfer else get_transfer_file_path_tooltip_text
+
+    return show_tooltip(treeview, pos_x, pos_y, tooltip, sourcecolumn,
+                        ("folder", "filename", "path"), function)
 
 
-def show_user_status_tooltip(treeview, x, y, tooltip, sourcecolumn):
-    return show_tooltip(treeview, x, y, tooltip, sourcecolumn, ("status",), get_user_status_tooltip_text)
+def show_user_status_tooltip(treeview, pos_x, pos_y, tooltip, sourcecolumn):
+    return show_tooltip(treeview, pos_x, pos_y, tooltip, sourcecolumn, ("status",), get_user_status_tooltip_text)
