@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2021 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Team
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2016 Mutnick <muhing@yahoo.com>
 # COPYRIGHT (C) 2008-2011 Quinox <quinox@users.sf.net>
@@ -44,11 +44,9 @@ class PluginHandler:
         self.core = core
         self.config = config
 
-        log.add_debug("Loading plugin handler")
-
-        self.my_username = self.config.sections["server"]["login"]
         self.plugindirs = []
         self.enabled_plugins = {}
+        self.command_source = None
 
         try:
             os.makedirs(config.plugin_dir)
@@ -62,20 +60,34 @@ class PluginHandler:
         # Load home directory plugins
         self.plugindirs.append(config.plugin_dir)
 
-        if os.path.isdir(config.plugin_dir):
-            self.load_enabled()
-        else:
-            log.add(_("It appears '%s' is not a directory, not loading plugins."), config.plugin_dir)
+        BasePlugin.parent = self
+        BasePlugin.config = self.config
+        BasePlugin.core = self.core
+        BasePlugin.frame = self.core.ui_callback
+
+        self.load_enabled()
+
+    def quit(self):
+
+        # Notify plugins
+        self.shutdown_notification()
+
+        # Disable plugins
+        for plugin in self.list_installed_plugins():
+            self.disable_plugin(plugin)
 
     def update_completions(self, plugin):
 
-        if plugin.__publiccommands__ and self.config.sections["words"]["commands"]:
+        if not self.config.sections["words"]["commands"]:
+            return
+
+        if plugin.__publiccommands__:
             self.core.chatrooms.update_completions()
 
-        if plugin.__privatecommands__ and self.config.sections["words"]["commands"]:
+        if plugin.__privatecommands__:
             self.core.privatechats.update_completions()
 
-    def __findplugin(self, plugin_name):
+    def findplugin(self, plugin_name):
         for directory in self.plugindirs:
             fullpath = os.path.join(directory, plugin_name)
 
@@ -102,26 +114,31 @@ class PluginHandler:
 
         except Exception:
             # Import user plugin
-            path = self.__findplugin(plugin_name)
+            path = self.findplugin(plugin_name)
 
             if path is None:
                 log.add(_("Failed to load plugin '%s', could not find it."), plugin_name)
-                return False
+                return None
+
+            # Add plugin folder to path in order to support relative imports
+            sys.path.append(path)
 
             import importlib.util
             spec = importlib.util.spec_from_file_location(plugin_name, os.path.join(path, '__init__.py'))
             plugin = importlib.util.module_from_spec(spec)
             spec.loader.exec_module(plugin)
 
-        instance = plugin.Plugin(self, self.config, self.core)
+        instance = plugin.Plugin()
+        instance.internal_name = BasePlugin.internal_name
+        instance.human_name = BasePlugin.human_name
+
         self.plugin_settings(plugin_name, instance)
 
         if hasattr(plugin, "enable"):
-            instance.log("top-level enable() function is obsolete, please use BasePlugin.init() instead" % plugin_name)
+            instance.log("top-level enable() function is obsolete, please use BasePlugin.__init__() instead")
 
         if hasattr(plugin, "disable"):
-            instance.log("top-level disable() function is obsolete, please use BasePlugin.disable() instead"
-                         % plugin_name)
+            instance.log("top-level disable() function is obsolete, please use BasePlugin.disable() instead")
 
         if hasattr(instance, "LoadNotification"):
             instance.log("LoadNotification() is obsolete, please use init()")
@@ -133,7 +150,7 @@ class PluginHandler:
         # Our config file doesn't play nicely with some characters
         if "=" in plugin_name:
             log.add(
-                _("Unable to enable plugin %(name)s. Plugin folder name contains invalid characters: %(characters)s"), {
+                _("Unable to load plugin %(name)s. Plugin folder name contains invalid characters: %(characters)s"), {
                     "name": plugin_name,
                     "characters": "="
                 })
@@ -143,10 +160,13 @@ class PluginHandler:
             return False
 
         try:
+            BasePlugin.internal_name = plugin_name
+            BasePlugin.human_name = self.get_plugin_info(plugin_name).get("Name", plugin_name)
+
             plugin = self.load_plugin(plugin_name)
 
-            if not plugin:
-                raise Exception("Error loading plugin '%s'" % plugin_name)
+            if plugin is None:
+                return False
 
             plugin.init()
 
@@ -159,11 +179,12 @@ class PluginHandler:
             self.update_completions(plugin)
 
             self.enabled_plugins[plugin_name] = plugin
-            log.add(_("Enabled plugin %s"), plugin.__name__)
+            plugin.loaded_notification()
+            log.add(_("Loaded plugin %s"), plugin.human_name)
 
         except Exception:
             from traceback import format_exc
-            log.add(_("Unable to enable plugin %(module)s\n%(exc_trace)s"),
+            log.add(_("Unable to load plugin %(module)s\n%(exc_trace)s"),
                     {'module': plugin_name, 'exc_trace': format_exc()})
             return False
 
@@ -184,10 +205,10 @@ class PluginHandler:
         if plugin_name not in self.enabled_plugins:
             return False
 
-        try:
-            plugin = self.enabled_plugins[plugin_name]
+        plugin = self.enabled_plugins[plugin_name]
+        path = self.findplugin(plugin_name)
 
-            log.add(_("Disabled plugin {}".format(plugin.__name__)))
+        try:
             plugin.disable()
 
             for trigger, _func in plugin.__publiccommands__:
@@ -197,16 +218,33 @@ class PluginHandler:
                 self.core.privatechats.CMDS.remove('/' + trigger + ' ')
 
             self.update_completions(plugin)
-
-            del self.enabled_plugins[plugin_name]
-            del plugin
+            plugin.unloaded_notification()
+            log.add(_("Unloaded plugin %s"), plugin.human_name)
 
         except Exception:
             from traceback import format_exc
-            log.add(_("Unable to fully disable plugin %(module)s\n%(exc_trace)s"),
+            log.add(_("Unable to unload plugin %(module)s\n%(exc_trace)s"),
                     {'module': plugin_name, 'exc_trace': format_exc()})
             return False
 
+        finally:
+            # Remove references to relative modules
+            if path in sys.path:
+                sys.path.remove(path)
+
+            for name, module in list(sys.modules.items()):
+                try:
+                    if module.__file__.startswith(path):
+                        sys.modules.pop(name, None)
+                        del module
+
+                except AttributeError:
+                    # Builtin module
+                    continue
+
+            del self.enabled_plugins[plugin_name]
+            del plugin
+
         return True
 
     def get_plugin_settings(self, plugin_name):
@@ -219,20 +257,47 @@ class PluginHandler:
         return None
 
     def get_plugin_info(self, plugin_name):
-        path = os.path.join(self.__findplugin(plugin_name), 'PLUGININFO')
 
-        with open(path) as file_handle:
-            infodict = {}
+        infodict = {}
+        plugin_folder = self.findplugin(plugin_name)
 
+        if plugin_folder is None:
+            return infodict
+
+        info_path = os.path.join(self.findplugin(plugin_name), 'PLUGININFO')
+
+        with open(info_path, encoding="utf-8") as file_handle:
             for line in file_handle:
                 try:
-                    key, val = line.split("=", 1)
-                    infodict[key.strip()] = literal_eval(val.strip())
-                except ValueError:
-                    pass  # this happens on blank lines
+                    key, value = line.split("=", 1)
+                    key = key.strip()
+                    value = value.strip()
+
+                    # Translatable string
+                    if value.startswith("_(") and value.endswith(")"):
+                        infodict[key] = _(literal_eval(value[2:-1]))
+                        continue
+
+                    infodict[key] = literal_eval(value)
+
+                except Exception:
+                    pass  # this can happen on blank lines
 
         return infodict
 
+    @staticmethod
+    def show_plugin_error(plugin_name, exc_type, exc_value, exc_traceback):
+
+        from traceback import format_tb
+
+        log.add(_("Plugin %(module)s failed with error %(errortype)s: %(error)s.\n"
+                  "Trace: %(trace)s"), {
+            'module': plugin_name,
+            'errortype': exc_type,
+            'error': exc_value,
+            'trace': ''.join(format_tb(exc_traceback))
+        })
+
     def save_enabled(self):
         self.config.sections["plugins"]["enabled"] = list(self.enabled_plugins.keys())
 
@@ -242,12 +307,16 @@ class PluginHandler:
         if not enable:
             return
 
+        log.add(_("Loading plugin system"))
+
         to_enable = self.config.sections["plugins"]["enabled"]
+        log.add_debug("Enabled plugin(s): %s" % ', '.join(to_enable))
 
         for plugin in to_enable:
             self.enable_plugin(plugin)
 
     def plugin_settings(self, plugin_name, plugin):
+        plugin_name = plugin_name.lower()
         try:
             if not plugin.settings:
                 return
@@ -266,10 +335,13 @@ class PluginHandler:
                     plugin.settings[key] = customsettings[key]
 
                 else:
-                    log.add(_("Stored setting '%(name)s' is no longer present in the plugin"), {'name': key})
+                    log.add_debug("Stored setting '%(key)s' is no longer present in the '%(name)s' plugin", {
+                        'key': key,
+                        'name': plugin_name
+                    })
 
         except KeyError:
-            log.add("No custom settings found for %s", (plugin.__name__,))
+            log.add_debug("No stored settings found for %s", plugin.human_name)
 
     def trigger_public_command_event(self, room, command, args):
         return self._trigger_command(command, room, args, public_command=True)
@@ -279,50 +351,39 @@ class PluginHandler:
 
     def _trigger_command(self, command, source, args, public_command):
 
+        self.command_source = (public_command, source)
+
         for module, plugin in self.enabled_plugins.items():
+            if plugin is None:
+                continue
+
+            return_value = None
+            commands = plugin.__publiccommands__ if public_command else plugin.__privatecommands__
+
             try:
-                if plugin is None:
-                    continue
-
-                return_value = None
-
-                if public_command:
-                    for trigger, func in plugin.__publiccommands__:
-                        if trigger == command:
-                            return_value = func(plugin, source, args)
-
-                else:
-                    for trigger, func in plugin.__privatecommands__:
-                        if trigger == command:
-                            return_value = func(plugin, source, args)
-
-                if return_value is None:
-                    # Nothing changed, continue to the next plugin
-                    continue
-
-                if return_value == returncode['zap']:
-                    return True
-
-                if return_value == returncode['pass']:
-                    continue
-
-                log.add(_("Plugin %(module)s returned something weird, '%(value)s', ignoring"),
-                        {'module': module, 'value': str(return_value)})
+                for trigger, func in commands:
+                    if trigger == command:
+                        return_value = getattr(plugin, func.__name__)(source, args)
 
             except Exception:
-                from traceback import extract_stack
-                from traceback import extract_tb
-                from traceback import format_list
+                self.show_plugin_error(module, sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
+                continue
 
-                log.add(_("Plugin %(module)s failed with error %(errortype)s: %(error)s.\n"
-                          "Trace: %(trace)s\nProblem area:%(area)s"), {
-                    'module': module,
-                    'errortype': sys.exc_info()[0],
-                    'error': sys.exc_info()[1],
-                    'trace': ''.join(format_list(extract_stack())),
-                    'area': ''.join(format_list(extract_tb(sys.exc_info()[2])))
-                })
+            if return_value is None:
+                # Nothing changed, continue to the next plugin
+                continue
 
+            if return_value == returncode['zap']:
+                self.command_source = None
+                return True
+
+            if return_value == returncode['pass']:
+                continue
+
+            log.add_debug("Plugin %(module)s returned something weird, '%(value)s', ignoring",
+                          {'module': module, 'value': str(return_value)})
+
+        self.command_source = None
         return False
 
     def trigger_event(self, function_name, args):
@@ -330,7 +391,7 @@ class PluginHandler:
         are precisely the same except for how n+ responds to them, both can be
         triggered by this function. """
 
-        function_name_camelcase = ''.join(x.capitalize() or '_' for x in function_name.split('_'))
+        function_name_camelcase = function_name.title().replace('_', '')
 
         for module, plugin in self.enabled_plugins.items():
             try:
@@ -343,54 +404,44 @@ class PluginHandler:
                 else:
                     return_value = getattr(plugin, function_name)(*args)
 
-                if return_value is None:
-                    # Nothing changed, continue to the next plugin
-                    continue
-
-                if isinstance(return_value, tuple):
-                    # The original args were modified, update them
-                    args = return_value
-                    continue
-
-                if return_value == returncode['zap']:
-                    return None
-
-                if return_value == returncode['break']:
-                    return args
-
-                if return_value == returncode['pass']:
-                    continue
-
-                log.add(_("Plugin %(module)s returned something weird, '%(value)s', ignoring"),
-                        {'module': module, 'value': return_value})
-
             except Exception:
-                from traceback import extract_stack
-                from traceback import extract_tb
-                from traceback import format_list
+                self.show_plugin_error(module, sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
+                continue
 
-                log.add(_("Plugin %(module)s failed with error %(errortype)s: %(error)s.\n"
-                          "Trace: %(trace)s\nProblem area:%(area)s"), {
-                    'module': module,
-                    'errortype': sys.exc_info()[0],
-                    'error': sys.exc_info()[1],
-                    'trace': ''.join(format_list(extract_stack())),
-                    'area': ''.join(format_list(extract_tb(sys.exc_info()[2])))
-                })
+            if return_value is None:
+                # Nothing changed, continue to the next plugin
+                continue
+
+            if isinstance(return_value, tuple):
+                # The original args were modified, update them
+                args = return_value
+                continue
+
+            if return_value == returncode['zap']:
+                return None
+
+            if return_value == returncode['break']:
+                return args
+
+            if return_value == returncode['pass']:
+                continue
+
+            log.add_debug("Plugin %(module)s returned something weird, '%(value)s', ignoring",
+                          {'module': module, 'value': return_value})
 
         return args
 
-    def search_request_notification(self, searchterm, user, searchid):
-        self.trigger_event("search_request_notification", (searchterm, user, searchid))
+    def search_request_notification(self, searchterm, user, token):
+        self.trigger_event("search_request_notification", (searchterm, user, token))
 
-    def distrib_search_notification(self, searchterm, user, searchid):
-        self.trigger_event("distrib_search_notification", (searchterm, user, searchid))
+    def distrib_search_notification(self, searchterm, user, token):
+        self.trigger_event("distrib_search_notification", (searchterm, user, token))
 
     def public_room_message_notification(self, room, user, line):
         self.trigger_event("public_room_message_notification", (room, user, line))
 
     def incoming_private_chat_event(self, user, line):
-        if user != self.my_username:
+        if user != self.core.login_username:
             # dont trigger the scripts on our own talking - we've got "Outgoing" for that
             return self.trigger_event("incoming_private_chat_event", (user, line))
 
@@ -460,6 +511,9 @@ class PluginHandler:
     def user_stats_notification(self, user, stats):
         self.trigger_event("user_stats_notification", (user, stats))
 
+    def user_status_notification(self, user, status, privileged):
+        self.trigger_event("user_status_notification", (user, status, privileged))
+
     def upload_queued_notification(self, user, virtual_path, real_path):
         self.trigger_event("upload_queued_notification", (user, virtual_path, real_path))
 
@@ -522,7 +576,7 @@ class ResponseThrottle:
 
         port = False
         try:
-            _ip_address, port = self.core.users[nick].addr
+            _ip_address, port = self.core.protothread.user_addresses[nick]
         except Exception:
             port = True
 
@@ -571,35 +625,49 @@ class ResponseThrottle:
 
 class BasePlugin:
 
-    __name__ = "BasePlugin"
     __publiccommands__ = []
     __privatecommands__ = []
     settings = {}
     metasettings = {}
 
-    def __init__(self, parent, config, core):
+    internal_name = None
+    human_name = None
+    parent = None
+    config = None
+    core = None
+    frame = None
 
-        # Never override this function, override init() instead
-        self.parent = parent
-        self.config = config
-        self.core = core
-        self.frame = core.ui_callback
+    def __init__(self):
+        # The plugin class is initializing, plugin settings are not available yet
+        pass
 
     def init(self):
-        # Custom enable function for plugins
+        # Called after __init__() when plugin settings have loaded
+        pass
+
+    def loaded_notification(self):
+        # The plugin has finished loaded (settings are loaded at this stage)
         pass
 
     def disable(self):
-        # Custom disable function for plugins
+        # The plugin has started unloading
+        pass
+
+    def unloaded_notification(self):
+        # The plugin has finished unloading
+        pass
+
+    def shutdown_notification(self):
+        # Application is shutting down
         pass
 
     def public_room_message_notification(self, room, user, line):
         pass
 
-    def search_request_notification(self, searchterm, user, searchid):
+    def search_request_notification(self, searchterm, user, token):
         pass
 
-    def distrib_search_notification(self, searchterm, user, searchid):
+    def distrib_search_notification(self, searchterm, user, token):
         pass
 
     def incoming_private_chat_event(self, user, line):
@@ -662,6 +730,9 @@ class BasePlugin:
     def user_stats_notification(self, user, stats):
         pass
 
+    def user_status_notification(self, user, status, privileged):
+        pass
+
     def upload_queued_notification(self, user, virtual_path, real_path):
         pass
 
@@ -677,33 +748,11 @@ class BasePlugin:
     def download_finished_notification(self, user, virtual_path, real_path):
         pass
 
-    def shutdown_notification(self):
-        pass
-
     # The following are functions to make your life easier,
     # you shouldn't override them.
-    def saypublic(self, room, text):
-        self.log("saypublic(room, text) is deprecated, please use send_public(room, text)")
-        self.send_public(room, text)
-
-    def sayprivate(self, user, text):
-        self.log("sayprivate(user, text) is deprecated, please use send_private(user, text)")
-        self.send_private(user, text, show_ui=True)
-
-    def sendprivate(self, user, text):
-        self.log("sendprivate(user, text) is deprecated, please use send_private(user, text, show_ui=False)")
-        self.send_private(user, text, show_ui=False)
-
-    def fakepublic(self, room, user, text):
-
-        self.log("fakepublic(room, user, text) is deprecated, please use echo_public(room, text)")
-
-        msg = slskmessages.SayChatroom(room, text)
-        msg.user = user
-        self.core.chatrooms.say_chat_room(msg)
 
     def log(self, msg, msg_args=None):
-        log.add(self.__name__ + ": " + msg, msg_args)
+        log.add(self.human_name + ": " + msg, msg_args)
 
     def send_public(self, room, text):
         self.core.queue.append(slskmessages.SayChatroom(room, text))
@@ -732,3 +781,43 @@ class BasePlugin:
 
         self.core.privatechats.show_user(user)
         self.core.privatechats.echo_message(user, text, message_type)
+
+    def send_message(self, text):
+        """ Convenience function to send a message to the same user/room
+        a plugin command runs for """
+
+        if self.parent.command_source is None:
+            # Function was not called from a command
+            return
+
+        public_command, source = self.parent.command_source
+        function = self.send_public if public_command else self.send_private
+
+        function(source, text)
+
+    def echo_message(self, text, message_type="local"):
+        """ Convenience function to display a raw message the same window
+        a plugin command runs from """
+
+        if self.parent.command_source is None:
+            # Function was not called from a command
+            return
+
+        public_command, source = self.parent.command_source
+        function = self.echo_public if public_command else self.echo_private
+
+        function(source, text, message_type)
+
+    # Obsolete functions
+
+    def saypublic(self, _room, _text):
+        self.log("saypublic(room, text) is obsolete, please use send_public(room, text)")
+
+    def sayprivate(self, _user, _text):
+        self.log("sayprivate(user, text) is obsolete, please use send_private(user, text)")
+
+    def sendprivate(self, _user, _text):
+        self.log("sendprivate(user, text) is obsolete, please use send_private(user, text, show_ui=False)")
+
+    def fakepublic(self, _room, _user, _text):
+        self.log("fakepublic(room, user, text) is obsolete, please use echo_public(room, text)")
