@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2021 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Team
 # COPYRIGHT (C) 2020 Lene Preuss <lene.preuss@gmail.com>
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2007 Daelstorm <daelstorm@gmail.com>
@@ -26,15 +26,22 @@ This module contains utility functions.
 """
 
 import errno
+import json
 import os
 import pickle
+import sys
+import webbrowser
 
 from pynicotine.config import config
 from pynicotine.logfacility import log
 
-ILLEGALPATHCHARS = ["?", ":", ">", "<", "|", "*", '"']
-ILLEGALFILECHARS = ILLEGALPATHCHARS + ["\\", "/"]
+FILE_SIZE_SUFFIXES = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
+PUNCTUATION = ['!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>',
+               '?', '@', '[', '\\', ']', '^', '_', '`', '{', '|', '}', '~', '–', '—', '‐', '’', '“', '”', '…']
+ILLEGALPATHCHARS = ['?', ':', '>', '<', '|', '*', '"']
+ILLEGALFILECHARS = ILLEGALPATHCHARS + ['\\', '/']
 REPLACEMENTCHAR = '_'
+OPEN_SOULSEEK_URL = None
 
 
 def rename_process(new_name, debug_info=False):
@@ -66,7 +73,7 @@ def rename_process(new_name, debug_info=False):
         msg = ["Errors occurred while trying to change process name:"]
         for i in errors:
             msg.append("%s" % (i,))
-        print('\n'.join(msg))
+        log.add('\n'.join(msg))
 
 
 def clean_file(filename):
@@ -92,8 +99,8 @@ def clean_path(path, absolute=False):
 
     path = ''.join([drive, path])
 
-    # Path can never end with a period on Windows machines
-    path = path.rstrip('.')
+    # Path can never end with a period or space on Windows machines
+    path = path.rstrip('. ')
 
     return path
 
@@ -116,13 +123,108 @@ def get_path(folder_name, base_name, callback, data=None):
         callback(filepath, data)
 
 
-def get_latest_version():
+def open_file_path(file_path, command=None):
+    """ Currently used to either open a folder or play an audio file
+    Tries to run a user-specified command first, and falls back to
+    the system default. """
 
-    import json
+    try:
+        file_path = os.path.normpath(file_path)
+
+        if command and "$" in command:
+            execute_command(command, file_path)
+
+        elif sys.platform == "win32":
+            os.startfile(file_path)
+
+        elif sys.platform == "darwin":
+            execute_command("open $", file_path)
+
+        elif not webbrowser.open(file_path):
+            raise webbrowser.Error("no known URL provider available")
+
+    except Exception as error:
+        log.add(_("Failed to open file path: %s"), error)
+
+
+def open_uri(uri):
+    """ Open a URI in an external (web) browser. The given argument has
+    to be a properly formed URI including the scheme (fe. HTTP). """
+
+    # Situation 1, user defined a way of handling the protocol
+    protocol = uri[:uri.find(":")]
+    protocol_handlers = config.sections["urls"]["protocols"]
+
+    if protocol in protocol_handlers and protocol_handlers[protocol]:
+        try:
+            execute_command(protocol_handlers[protocol], uri)
+            return True
+
+        except RuntimeError as error:
+            log.add(error)
+
+    if protocol == "slsk":
+        OPEN_SOULSEEK_URL(uri.strip())  # pylint:disable=not-callable
+        return True
+
+    # Situation 2, user did not define a way of handling the protocol
+    try:
+        if not webbrowser.open(uri):
+            raise webbrowser.Error("no known URL provider available")
+
+        return True
+
+    except Exception as error:
+        log.add(_("Failed to open URL: %s"), error)
+
+    return False
+
+
+def open_log(folder, filename):
+    _handle_log(folder, filename, open_log_callback)
+
+
+def delete_log(folder, filename):
+    _handle_log(folder, filename, delete_log_callback)
+
+
+def _handle_log(folder, filename, callback):
+
+    try:
+        if not os.path.isdir(folder):
+            os.makedirs(folder)
+
+        filename = clean_file(filename) + ".log"
+        get_path(folder, filename, callback)
+
+    except Exception as error:
+        log.add("Failed to process log file: %s", error)
+
+
+def open_log_callback(path, _data):
+
+    if not os.path.exists(path):
+        with open(path, "w", encoding="utf-8"):
+            # No logs, create empty file
+            pass
+
+    open_file_path(path)
+
+
+def delete_log_callback(path, _data):
+
+    with open(path, "w", encoding="utf-8"):
+        # Check if path should contain special characters
+        pass
+
+    os.remove(path)
+
+
+def get_latest_version():
 
     response = http_request(
         "https", "pypi.org", "/pypi/nicotine-plus/json",
-        headers={"User-Agent": "Nicotine+"}
+        headers={"User-Agent": config.application_name}
     )
     data = json.loads(response)
 
@@ -150,6 +252,22 @@ def make_version(version):
     return (major << 24) + (minor << 16) + (patch << 8) + stable
 
 
+def human_length(seconds):
+
+    minutes, seconds = divmod(seconds, 60)
+    hours, minutes = divmod(minutes, 60)
+    days, hours = divmod(hours, 24)
+
+    if days > 0:
+        ret = '%i:%02i:%02i:%02i' % (days, hours, minutes, seconds)
+    elif hours > 0:
+        ret = '%i:%02i:%02i' % (hours, minutes, seconds)
+    else:
+        ret = '%i:%02i' % (minutes, seconds)
+
+    return ret
+
+
 def get_result_bitrate_length(filesize, attributes):
     """ Used to get the audio bitrate and length of search results and
     user browse files """
@@ -168,6 +286,7 @@ def get_result_bitrate_length(filesize, attributes):
         third = attributes[2]
 
         # Sometimes the vbr indicator is in third position
+        # Known clients: Soulseek NS, Nicotine+, Museek+, SoulSeeX
         if third in (0, 1):
 
             if third == 1:
@@ -177,9 +296,10 @@ def get_result_bitrate_length(filesize, attributes):
             h_bitrate = str(bitrate) + h_bitrate
 
             length = second
-            h_length = '%i:%02i' % (second / 60, second % 60)
+            h_length = human_length(second)
 
         # Sometimes the vbr indicator is in second position
+        # Known clients: unknown (does this actually exist?)
         elif second in (0, 1):
 
             if second == 1:
@@ -189,31 +309,34 @@ def get_result_bitrate_length(filesize, attributes):
             h_bitrate = str(bitrate) + h_bitrate
 
             length = third
-            h_length = '%i:%02i' % (third / 60, third % 60)
+            h_length = human_length(third)
 
         # Lossless audio, length is in first position
+        # Known clients: SoulseekQt 2015-6-12 and later
         elif third > 1:
 
+            length = first
+            h_length = human_length(first)
+
             # Bitrate = sample rate (Hz) * word length (bits) * channel count
             # Bitrate = 44100 * 16 * 2
-            bitrate = (second * third * 2) / 1000
+            bitrate = (second * third * 2) // 1000
             h_bitrate = str(bitrate)
 
-            length = first
-            h_length = '%i:%02i' % (first / 60, first % 60)
-
         else:
 
             bitrate = first
             h_bitrate = str(bitrate) + h_bitrate
 
     # If there are 2 entries in the attribute list
+    # Known clients: SoulseekQt
     elif len(attributes) == 2:
 
         first = attributes[0]
         second = attributes[1]
 
         # Sometimes the vbr indicator is in second position
+        # Known clients: SoulseekQt 2015-2-21 and earlier
         if second in (0, 1):
 
             # If it's a vbr file we can't deduce the length
@@ -230,89 +353,76 @@ def get_result_bitrate_length(filesize, attributes):
                 bitrate = first
                 h_bitrate = str(bitrate) + h_bitrate
 
-                if bitrate <= 0:
-                    length = 0
-                else:
+                if bitrate > 0:
                     # Dividing the file size by the bitrate in Bytes should give us a good enough approximation
-                    length = filesize / (bitrate / 8 * 1000)
+                    length = filesize / (bitrate * 125)
+                    h_length = human_length(length)
 
-                h_length = '%i:%02i' % (length / 60, length % 60)
+        # Lossless audio without length attribute
+        # Known clients: SoulseekQt 2015-6-12 and later
+        elif first >= 8000 and second <= 64:
+
+            # Bitrate = sample rate (Hz) * word length (bits) * channel count
+            # Bitrate = 44100 * 16 * 2
+            bitrate = (first * second * 2) // 1000
+            h_bitrate = str(bitrate)
+
+            if bitrate > 0:
+                # Dividing the file size by the bitrate in Bytes should give us a good enough approximation
+                length = filesize / (bitrate * 125)
+                h_length = human_length(length)
 
         # Sometimes the bitrate is in first position and the length in second position
+        # Known clients: SoulseekQt 2015-6-12 and later
         else:
 
             bitrate = first
             h_bitrate = str(bitrate) + h_bitrate
 
             length = second
-            h_length = '%i:%02i' % (second / 60, second % 60)
+            h_length = human_length(second)
+
+    # Ignore invalid values
+    if bitrate <= 0:
+        h_bitrate = ""
+        bitrate = 0
+
+    if length < 0:
+        h_length = ""
+        length = 0
 
     return h_bitrate, bitrate, h_length, length
 
 
-size_suffixes = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
+def _human_speed_or_size(unit):
+
+    template = "%.3g %s"
+    try:
+        for suffix in FILE_SIZE_SUFFIXES:
+            if unit < 1024:
+                if unit > 999:
+                    template = "%.4g %s"
+
+                return template % (unit, suffix)
+
+            unit /= 1024
+
+    except TypeError:
+        pass
+
+    return unit
+
+
+def human_speed(speed):
+    return _human_speed_or_size(speed) + "/s"
 
 
 def human_size(filesize):
-    try:
-        step_unit = 1024.0
-
-        for i in size_suffixes:
-            if filesize < step_unit:
-                return "%3.1f %s" % (filesize, i)
-
-            filesize /= step_unit
-
-    except TypeError:
-        pass
-
-    return filesize
-
-
-speed_suffixes = ['B/s', 'KiB/s', 'MiB/s', 'GiB/s', 'TiB/s', 'PiB/s', 'EiB/s', 'ZiB/s', 'YiB/s']
-
-
-def human_speed(filesize):
-    try:
-        step_unit = 1024.0
-
-        for i in speed_suffixes:
-            if filesize < step_unit:
-                return "%3.1f %s" % (filesize, i)
-
-            filesize /= step_unit
-
-    except TypeError:
-        pass
-
-    return filesize
+    return _human_speed_or_size(filesize)
 
 
 def humanize(number):
-
-    fashion = config.sections["ui"]["decimalsep"]
-
-    if fashion in ("", "<None>"):
-        return str(number)
-
-    if fashion == "<space>":
-        fashion = " "
-
-    number = str(number)
-
-    if number[0] == "-":
-        neg = "-"
-        number = number[1:]
-    else:
-        neg = ""
-
-    ret = ""
-
-    while number[-3:]:
-        part, number = number[-3:], number[:-3]
-        ret = "%s%s%s" % (part, fashion, ret)
-
-    return neg + ret[:-1]
+    return "{:n}".format(number)
 
 
 def unescape(string):
@@ -366,8 +476,8 @@ def execute_command(command, replacement=None, background=True, returnoutput=Fal
     if command.endswith("&"):
         command = command[:-1]
         if returnoutput:
-            print("Yikes, I was asked to return output but I'm also asked to launch "
-                  "the process in the background. returnoutput gets precedent.")
+            log.add("Yikes, I was asked to return output but I'm also asked to launch "
+                    "the process in the background. returnoutput gets precedent.")
         else:
             background = True
 
@@ -578,7 +688,7 @@ def add_alias(rest):
 
     msg = "\n" + _("Aliases:") + "\n"
 
-    for (key, value) in aliases.items():
+    for key, value in aliases.items():
         msg = msg + "%s: %s\n" % (key, value)
 
     return msg + "\n"
@@ -740,7 +850,7 @@ def debug(*args):
     """ Prints debugging info. """
 
     truncated_args = [arg[:200] if isinstance(arg, str) else arg for arg in args]
-    print('*' * 8, truncated_args)
+    log.add('*' * 8, truncated_args)
 
 
 def strace(function):
@@ -750,9 +860,9 @@ def strace(function):
 
     def newfunc(*args, **kwargs):
         name = function.__name__
-        print(("%s(%s)" % (name, ", ".join(map(repr, chain(args, list(kwargs.values())))))))
+        log.add("%s(%s)" % (name, ", ".join(map(repr, chain(args, list(kwargs.values()))))))
         retvalue = function(*args, **kwargs)
-        print(("%s(%s): %s" % (name, ", ".join(map(repr, chain(args, list(kwargs.values())))), repr(retvalue))))
+        log.add("%s(%s): %s" % (name, ", ".join(map(repr, chain(args, list(kwargs.values())))), repr(retvalue)))
         return retvalue
 
     return newfunc
